#
# Copyright 2019 Venafi, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import absolute_import, division, generators, unicode_literals, print_function, nested_scopes, \
    with_statement

import datetime
import logging as log
import socket
import sys

import ipaddress
from builtins import bytes
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID, ExtensionOID
from six import string_types, binary_type

from .errors import VenafiConnectionError, ServerUnexptedBehavior, BadData, ClientBadData
from .http import HTTPStatus
from .policy import PolicySpecification


MIME_JSON = "application/json"
MIME_HTML = "text/html"
MIME_TEXT = "text/plain"
MIME_CSV = "text/csv"
MIME_ANY = "*/*"
LOCALHOST = "127.0.0.1"


class CertField:
    def __init__(self, value, locked=False):
        self.value = value
        self.locked = locked

    def __bool__(self):
        return bool(self.value)


def log_errors(data):
    if not isinstance(data, dict) or "errors" not in data:
        log.error("Unknown error format: %s", data)
        return
    for e in data["errors"]:
        log.error("%s: %s" % (e['code'], e['message']))


def get_ip_address():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    ip = None
    try:
        # doesn't even have to be reachable, it just allows the socket to be opened
        s.connect(('10.255.255.255', 1))
        ip = s.getsockname()[0]
    except Exception:
        ip = LOCALHOST
    finally:
        s.close()
    return ip


class KeyType:
    ALLOWED_SIZES = [2048, 3072, 4096, 8192]
    ALLOWED_CURVES = ["p256", "p384", "p521"]
    RSA = "rsa"
    ECDSA = "ec"

    def __init__(self, key_type, option):
        """
        :param str key_type:
        :param Union[str, int] option: key length for RSA (in int) or curve name for ECDSA
        """
        self.key_type = {"rsa": "rsa", "ec": "ec", "ecdsa": "ec"}.get(key_type.lower().strip())
        if self.key_type == KeyType.RSA:
            if option not in KeyType.ALLOWED_SIZES:
                log.error("unknown size: %s" % option)
                raise BadData
        elif self.key_type == KeyType.ECDSA:
            option = {"secp521r1": "p521", "secp384r1": "p384", "secp256r1": "p256", "p256": "p256", "p384": "p384",
                      "p521": "p521"}[option.lower().strip()]
            if option not in KeyType.ALLOWED_CURVES:
                log.error("unknown curve: %s, should be one of %s" % (option, KeyType.ALLOWED_CURVES))
                raise BadData
        else:
            log.error("unknown key type: %s" % key_type)
            raise BadData
        self.option = option

    def __repr__(self):
        return "KeyType(%s, %s)" % (self.key_type, self.option)

    def __eq__(self, other):
        if not isinstance(other, KeyType):
            return False
        return self.key_type == other.key_type and self.option == other.option


class ZoneConfig:
    def __init__(self, organization, organizational_unit, country, province, locality,
                 policy, key_type):
        """
        :param CertField organization:
        :param CertField organizational_unit:
        :param CertField country:
        :param CertField province:
        :param CertField locality:
        :param Policy policy:
        :param KeyType key_type:
        """
        self.organization = organization
        self.organizational_unit = organizational_unit
        self.country = country
        self.province = province
        self.locality = locality
        self.policy = policy
        self.key_type = key_type


class TokenInfo:
    def __init__(self, access_token, expires, refresh_token):
        self.access_token = access_token
        self.expires = expires
        self.refresh_token = refresh_token


class Policy:
    # This is the Certificate Issuing Template.
    # It is used by both TPP and Cloud implementations, hence the name was adapted to something more generic.
    def __init__(self, policy_id=None, company_id=None, name=None, system_generated=None,
                 creation_date=None, subject_cn_regexes=None, subject_o_regexes=None,
                 subject_ou_regexes=None, subject_st_regexes=None, subject_l_regexes=None, subject_c_regexes=None,
                 san_regexes=None, key_types=None, key_reuse=None, cert_authority=None, cert_authority_account_id=None,
                 cert_authority_product_option_id=None, priority=None, modification_date=None, status=None, reason=None,
                 validity_period=None, recommended_settings=None):
        """
        :param str policy_id:
        :param str company_id:
        :param str name:
        :param bool system_generated:
        :param datetime.datetime creation_date:
        :param list[str] subject_cn_regexes:
        :param list[str] subject_o_regexes:
        :param list[str] subject_ou_regexes:
        :param list[str] subject_st_regexes:
        :param list[str] subject_l_regexes:
        :param list[str] subject_c_regexes:
        :param list[str] san_regexes:
        :param list[KeyType] key_types:
        :param bool key_reuse:
        :param str cert_authority:
        :param str cert_authority_account_id:
        :param str cert_authority_product_option_id:
        :param int priority:
        :param str modification_date:
        :param str status:
        :param str reason:
        :param str validity_period:
        :param RecommendedSettings recommended_settings:
        """
        self.id = policy_id
        self.company_id = company_id
        self.name = name
        self.system_generated = system_generated
        self.creation_date = creation_date
        self.SubjectCNRegexes = subject_cn_regexes
        self.SubjectORegexes = subject_o_regexes
        self.SubjectOURegexes = subject_ou_regexes
        self.SubjectSTRegexes = subject_st_regexes
        self.SubjectLRegexes = subject_l_regexes
        self.SubjectCRegexes = subject_c_regexes
        self.SANRegexes = san_regexes
        self.key_types = key_types
        self.key_reuse = key_reuse

        self.cert_authority = cert_authority
        self.cert_authority_account_id = cert_authority_account_id
        self.cert_authority_product_option_id = cert_authority_product_option_id
        self.priority = priority
        self.modification_date = modification_date
        self.status = status
        self.reason = reason
        self.validity_period = validity_period
        self.recommended_settings = recommended_settings

    def __repr__(self):
        return "Policy:\n" + "\n".join(["  %s: %s" % (k, v) for k, v in (
            ("Id", self.id),
            ("Name", self.name),
            ("KeyReuse", self.key_reuse),
            ("Created", self.creation_date)
        )])

    def __str__(self):
        return self.name


class RecommendedSettings:
    def __init__(self, subject_o_value=None, subject_ou_value=None, subject_l_value=None, subject_st_value=None,
                 subject_c_value=None, key_type=None, key_reuse=None):
        """
        :param str subject_o_value:
        :param str subject_ou_value:
        :param str subject_l_value:
        :param str subject_st_value:
        :param str subject_c_value:
        :param KeyType key_type:
        :param bool key_reuse:
        """
        self.subjectOValue = subject_o_value
        self.subjectOUValue = subject_ou_value
        self.subjectLValue = subject_l_value
        self.subjectSTValue = subject_st_value
        self.subjectCValue = subject_c_value
        self.keyType = key_type
        self.keyReuse = key_reuse


class CertificateRequest:
    def __init__(self, cert_id=None,
                 san_dns=None,
                 email_addresses="",
                 ip_addresses=None,
                 user_principal_names=None,
                 uniform_resource_identifiers=None,
                 attributes=None,
                 key_type=None,
                 private_key=None,
                 key_password=None,
                 csr=None,
                 friendly_name=None,
                 common_name=None,
                 thumbprint=None,
                 organization=None,
                 organizational_unit=None,
                 country=None,
                 province=None,
                 locality=None,
                 origin=None,
                 custom_fields=None,
                 timeout=0
                 ):
        """
        :param str cert_id: Certificate request id. Generating by server.
        :param list[str] san_dns: Alternative names for SNI.
        :param list[str] email_addresses: List of email addresses
        :param list[str] ip_addresses: List of IP addresses
        :param list[str] user_principal_names: user principal names in username@domain format
        :param list[str] uniform_resource_identifiers: URIs
        :param attributes:
        :param KeyType key_type: Type of asymmetric cryptography algorithm. Default is RSA 2048.
        :param asymmetric.PrivateKey private_key: String with pem encoded private key or  asymmetric.PrivateKey
        :param str key_password: Password for encrypted private key. Not supported at this moment.
        :param str csr: Certificate Signing Request in pem format
        :param str friendly_name: Name for certificate in the platform. If not specified common name will be used.
        :param str common_name: Common name of certificate. Usually domain name.
        :param str thumbprint: Certificate thumbprint. Can be used for identifying certificate on the platform.
        :param str origin: application identifier
        :param list[CustomField] custom_fields: list of custom fields values to be added to the certificate.
        :param int timeout: Timeout for the certificate to be retrieved from server. Measured in seconds.
        """

        self.chain_option = "last"
        self.san_dns = san_dns or []
        self.email_addresses = email_addresses
        self.ip_addresses = ip_addresses or []
        self.user_principal_names = user_principal_names or []
        self.uniform_resource_identifiers = uniform_resource_identifiers or []
        self.attributes = attributes
        self.key_password = key_password
        self.key_type = key_type
        self.private_key = private_key
        self.public_key = None
        self._public_key_from_private()
        self.friendly_name = friendly_name or common_name
        self.id = cert_id
        self.common_name = common_name
        self.thumbprint = thumbprint
        self.organization = organization
        self.organizational_unit = organizational_unit
        self.country = country
        self.province = province
        self.locality = locality
        # CSR should be last, because it checks subject to match with over parameters
        self.csr = csr
        self.origin = origin or "Venafi VCert-Python"
        self.custom_fields = custom_fields
        self.cert_guid = None
        self.timeout = timeout

    def __setattr__(self, key, value):
        if key == "key_password":
            if isinstance(value, string_types):
                value = value.encode()
        elif key == "common_name":
            if isinstance(value, binary_type):
                value = value.decode()
        elif key == "key_type":
            if value is not None and not isinstance(value, KeyType):
                raise ClientBadData("key_type should be instance of vcert.KeyType")
        elif key == "private_key":
            if isinstance(value, string_types):
                value = serialization.load_pem_private_key(value.encode(),
                                                           password=self.key_password, backend=default_backend())
            if isinstance(value, rsa.RSAPrivateKey):
                self.key_type = KeyType(KeyType.RSA, value.key_size)
            elif isinstance(value, ec.EllipticCurvePrivateKey):
                self.key_type = KeyType(KeyType.ECDSA, value.curve.name)
            elif value is None:
                self.public_key = None
            else:
                raise ClientBadData("invalid private key type %s" % type(value))
        elif key == "csr":
            if isinstance(value, binary_type):
                value = value.decode()
            elif not (isinstance(value, string_types) or value is None):
                raise ClientBadData("invalid csr type %s" % type(value))
            if value:
                csr = x509.load_pem_x509_csr(value.encode(), default_backend())
                cn = csr.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value
                if self.common_name and self.common_name != cn:
                    raise ClientBadData("Common name from CSR doesn`t matches to CertificateRequest.common_name")
                ips = []
                dns = []
                try:
                    for e in csr.extensions.get_extension_for_oid(ExtensionOID.SUBJECT_ALTERNATIVE_NAME).value:
                        if isinstance(e, x509.general_name.DNSName):
                            dns.append(e.value)
                        elif isinstance(e, x509.general_name.IPAddress):
                            ips.append(e.value.exploded)
                    if self.ip_addresses and sorted(self.ip_addresses) != sorted(ips):
                        raise ClientBadData
                    if self.san_dns and sorted(self.san_dns) != sorted(dns):
                        raise ClientBadData
                except x509.extensions.ExtensionNotFound:
                    pass
        self.__dict__[key] = value

    def _gen_key(self):
        if self.key_type == None:
            self.key_type = KeyType(KeyType.RSA, 2048)
        if self.key_type.key_type == KeyType.RSA:
            self.private_key = rsa.generate_private_key(
                public_exponent=65537,
                key_size=self.key_type.option,
                backend=default_backend()
            )
        elif self.key_type.key_type == KeyType.ECDSA:
            if self.key_type.option == "p521":
                curve = ec.SECP521R1()
            elif self.key_type.option == "p384":
                curve = ec.SECP384R1()
            elif self.key_type.option == "p256":
                curve = ec.SECP256R1()
            else:
                curve = ec.SECP256R1()
            self.private_key = ec.generate_private_key(
                curve, default_backend()
            )
        else:
            raise ClientBadData
        self._public_key_from_private()

    def build_csr(self):
        if not self.private_key:
            self._gen_key()

        csr_builder = x509.CertificateSigningRequestBuilder()
        subject = [x509.NameAttribute(NameOID.COMMON_NAME, self.common_name,)]
        if self.locality:
            subject.append(x509.NameAttribute(NameOID.LOCALITY_NAME, self.locality))
        if self.province:
            subject.append(x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, self.province))
        if self.country:
            subject.append(x509.NameAttribute(NameOID.COUNTRY_NAME, self.country))
        if self.organization:
            subject.append(x509.NameAttribute(NameOID.ORGANIZATION_NAME, self.organization))
        if self.organizational_unit:
            if isinstance(self.organizational_unit, string_types):
                subject.append(x509.NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, self.organizational_unit))
            elif isinstance(self.organizational_unit, list):
                for u in self.organizational_unit:
                    subject.append(x509.NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, u))

        csr_builder = csr_builder.subject_name(x509.Name(subject))

        alt_names = []
        if self.ip_addresses:
            for ip in self.ip_addresses:
                alt_names.append(x509.IPAddress(ipaddress.IPv4Address(ip)))

        if self.san_dns:
            for ns in self.san_dns:
                alt_names.append(x509.DNSName(ns))

        if self.email_addresses:
            for mail in self.email_addresses:
                alt_names.append(x509.RFC822Name(mail))

        if self.uniform_resource_identifiers:
            for uri in self.uniform_resource_identifiers:
                alt_names.append(x509.UniformResourceIdentifier(uri))

        if self.user_principal_names:
            for upn in self.user_principal_names:
                # Python cryptography library doesn't include
                # OID for UPN in any of the OID constants
                # See http://www.oid-info.com/get/1.3.6.1.4.1.311.20.2.3
                # and https://docs.microsoft.com/en-us/windows/security/identity-protection/smart-cards/smart-card-certificate-requirements-and-enumeration
                UPNOID = x509.ObjectIdentifier('1.3.6.1.4.1.311.20.2.3')
                # x509 library expects DER encoded string, so encode UPN into bytes
                # with ASN1 syntax for UTF8String, which is a type/length/value encoding
                # per https://docs.microsoft.com/en-us/windows/win32/seccertenroll/about-utf8string
                # Construct the array of bytes (note bytes are strings in python2)
                # by inserting the header consisting of the tag '0x0C' followed by the length of the upn
                if sys.version_info > (3, 0):
                    # For python3, since we have native bytes available we'll convert the string into bytes.
                    # However, for cases when this method is called with inputs extracted from existing certificates
                    # & csrs, the input will already be encoded as bytes. So we'll check the input, if it is a 
                    # string we'll convert it into a bytes object then insert our header. Otherwise, we'll just
                    # insert the header in the passed in bytes.
                    if isinstance(upn, str):
                        bupn = bytes(upn,'utf-8')
                    else:
                        bupn = upn
                    values = [12,len(upn)]
                    header = bytes(values)
                    data = header + bupn
                    alt_names.append(x509.OtherName(UPNOID, data))
                else:
                    # For python2, since there is no native bytes method, we'll use the bytes method 
                    # in the future module to convert the string we constructed into bytes.
                    bupn = ''.join(chr(x) for x in [12,len(upn)])
                    bupn = bupn + str(upn)
                    alt_names.append(x509.OtherName(UPNOID,bytes(bupn, 'utf-8')))

        csr_builder = csr_builder.add_extension(
            x509.SubjectAlternativeName(alt_names),
            critical=False,
        )

        csr_builder = csr_builder.sign(self.private_key, hashes.SHA256(),
                                       default_backend())
        self.csr = csr_builder.public_bytes(serialization.Encoding.PEM).decode()
        return

    @property
    def private_key_pem(self):
        if self.key_password:
            encryption = serialization.BestAvailableEncryption(self.key_password)
        else:
            encryption = serialization.NoEncryption()

        return self.private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=encryption,
        ).decode()

    def _public_key_from_private(self):
        if self.private_key is None:
            return
        self.public_key = self.private_key.public_key()

    @property
    def public_key_pem(self):
        return self.public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        ).decode()

    def update_from_zone_config(self, zone):
        # TODO: check request against zone configuration
        """
        :param ZoneConfig zone:
        """
        if zone.organization.locked or (not self.organization and zone.organization):
            self.organization = zone.organization.value
        if zone.organizational_unit.locked or (not self.organizational_unit and zone.organizational_unit):
            self.organizational_unit = zone.organizational_unit.value
        if zone.country.locked or (not self.country and zone.country):
            self.country = zone.country.value
        if zone.province.locked or (not self.province and zone.province):
            self.province = zone.province.value
        if zone.locality.locked or (not self.locality and zone.locality):
            self.locality = zone.locality.value
        if not self.key_type and zone.key_type:
            self.key_type = zone.key_type


class CustomField:
    def __init__(self, name, value):
        """
        :param str name: The name of the custom field
        :param str value: The value to be added to the specified custom field
        """
        self.name = name
        self.value = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value


class RevocationRequest:
    class RevocationReasons:
        NoReason = 0
        key_compromise = 1
        ca_compromise = 2
        affiliation_changed = 3
        superseded = 4
        cessation_of_operation = 5  # OriginalUseNoLongerValid

    def __init__(self, req_id=None, thumbprint=None,  reason=RevocationReasons.NoReason,
                 comments="Revoked via api with python bindings", disable=True):
        """
        :param req_id:
        :param thumbprint:
        """
        self.id = req_id
        self.thumbprint = thumbprint
        self.reason = reason
        self.comments = comments
        self.disable = disable


CLIENT_ID = "vcert-sdk"  # type: str
# Certificate Management scope
SCOPE_CM = "certificate:manage,revoke"  # type: str
# Policy Management scope
SCOPE_PM = "certificate:manage;configuration:manage"  # type: str
# Full Management scope
SCOPE_FULL = "certificate:manage,revoke;configuration:manage"  # type: str


class Authentication:
    def __init__(self, user=None, password=None, access_token=None, refresh_token=None, api_key=None, state=None,
                 token_expires=None, client_id=CLIENT_ID, scope=SCOPE_CM):
        self.user = user
        self.password = password
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.token_expires = token_expires
        self.api_key = api_key
        self.client_id = client_id
        self.scope = scope
        self.state = state


class AppDetails:
    def __init__(self, app_id=None, cit_map=None, company_id=None, name=None, description=None,
                 owner_ids_and_types=None, fq_dns=None, internal_fq_dns=None, external_ip_ranges=None,
                 internal_ip_ranges=None, internal_ports=None, fully_qualified_domain_names=None, ip_ranges=None,
                 ports=None, org_unit_id=None):
        """
        :param str app_id:
        :param dict cit_map:
        """
        self.app_id = app_id
        self.cit_alias_id_map = cit_map
        self.company_id = company_id
        self.name = name
        self.description = description
        self.owner_ids_and_types = owner_ids_and_types
        self.fq_dns = fq_dns
        self.internal_fq_dns = internal_fq_dns
        self.external_ip_ranges = external_ip_ranges
        self.internal_ip_ranges = internal_ip_ranges
        self.internal_ports = internal_ports
        self.fully_qualified_domain_names = fully_qualified_domain_names
        self.ip_ranges = ip_ranges
        self.ports = ports
        self.org_unit_id = org_unit_id


class CommonConnection:

    def auth(self):
        """
        Authorize connection on platform. Return user object.
        Optional for making calls. Connection controls auth status by itself.
        """
        raise NotImplementedError

    def request_cert(self, request, zone):
        """
        Making request to certificate. It will generate CSR from data if CSR not specified, generate key if required and send to server for signing. Set request.id for retrieving certificate.

        :param CertificateRequest request: Certificate in PEM format
        :param str zone: Venafi zone tag name
        :rtype: bool
        """
        raise NotImplementedError

    def retrieve_cert(self, request):
        """
        Get signed certificate from server by request.id

        :param CertificateRequest request:
        :rtype: Certificate
        """
        raise NotImplementedError

    def revoke_cert(self, request):
        """
        :param RevocationRequest request:
        """
        raise NotImplementedError

    def renew_cert(self, request, reuse_key=False):
        """
        :param CertificateRequest request:
        :param bool reuse_key:
        """
        raise NotImplementedError

    def read_zone_conf(self, tag):
        """
        :param str tag:
        :rtype: ZoneConfig
        """
        raise NotImplementedError

    def import_cert(self, request):
        raise NotImplementedError

    def get_policy(self, zone):
        """
        :param str zone:
        :rtype: PolicySpecification
        """
        raise NotImplementedError

    def set_policy(self, zone, policy_spec):
        """
        :param str zone:
        :param PolicySpecification policy_spec:
        :rtype: None
        """
        raise NotImplementedError

    @staticmethod
    def process_server_response(r):
        if r.status_code not in (HTTPStatus.OK, HTTPStatus.ACCEPTED, HTTPStatus.CREATED, HTTPStatus.CONFLICT):
            try:
                log_errors(r.json())
            except:
                log_errors(r.content)
            raise VenafiConnectionError("Server status: %s\nResponse: %s" % (r.status_code, r.request.url))

        content_type = r.headers.get("content-type")
        # Content-type not present, return status and reason (if any)
        if content_type is None:
            return r.status_code, r.reason
        if content_type.startswith(MIME_TEXT):
            log.debug(r.text)
            return r.status_code, r.text
        elif content_type.startswith(MIME_HTML):
            log.debug(r.text)
            return r.status_code, r.text
        # content-type in response is  application/json; charset=utf-8
        elif content_type.startswith(MIME_JSON):
            # log.debug(r.content.decode())
            return r.status_code, r.json()
        elif content_type.startswith(MIME_CSV):
            log.debug(r.content.decode())
            return r.status_code, r.content.decode()
        else:
            log.error("Unexpected content type: %s for request %s" % (content_type, r.request.url))
            raise ServerUnexptedBehavior



#
# Copyright 2020 Venafi, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from __future__ import (absolute_import, division, generators, unicode_literals, print_function, nested_scopes,
                        with_statement)

import base64
import logging as log
import re
import time

import requests
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.x509 import SignatureAlgorithmOID as AlgOID

from .common import MIME_JSON, TokenInfo, Authentication, KeyType, Policy, ZoneConfig, CertField
from .connection_tpp_abstract import AbstractTPPConnection, URLS
from .errors import (ClientBadData, ServerUnexptedBehavior, AuthenticationError, CertificateRequestError,
                     CertificateRenewError, VenafiError, RetrieveCertificateTimeoutError)
from .http import HTTPStatus
from .pem import parse_pem

HEADER_AUTHORIZATION = "Authorization"  # type: str

KEY_ACCESS_TOKEN = "access_token"  # type: str  # nosec
KEY_REFRESH_TOKEN = "refresh_token"  # type: str  # nosec
KEY_EXPIRATION_DATE = "expiration_date"  # type: str


class TPPTokenConnection(AbstractTPPConnection):
    def __init__(self, url, user=None, password=None, access_token=None, refresh_token=None, http_request_kwargs=None):
        """
        :param str url:
        :param str user:
        :param str password:
        :param str access_token:
        :param str refresh_token:
        :param dict[str,Any] http_request_kwargs:
        """
        self._base_url = url  # type: str
        self._auth = Authentication(user=user, password=password, access_token=access_token,
                                    refresh_token=refresh_token)  # type: Authentication
        if http_request_kwargs is None:
            http_request_kwargs = {"timeout": 180}
        elif "timeout" not in http_request_kwargs:
            http_request_kwargs["timeout"] = 180
        self._http_request_kwargs = http_request_kwargs or {}

    def __setattr__(self, key, value):
        if key == "_base_url":
            value = self._normalize_and_verify_base_url(value)
        self.__dict__[key] = value

    def __str__(self):
        return "[TPP] %s" % self._base_url

    def _get(self, url=None, params=None, check_token=True, include_headers=True):
        if check_token:
            self._check_token()

        headers = {}
        if include_headers:
            token = self._get_auth_header_value(self._auth.access_token)
            headers = {HEADER_AUTHORIZATION: token, 'content-type': MIME_JSON, 'cache-control': 'no-cache'}

        r = requests.get(self._base_url + url, headers=headers, params=params, **self._http_request_kwargs)
        return self.process_server_response(r)

    def _post(self, url=None, data=None, check_token=True, include_headers=True):
        if check_token:
            self._check_token()

        headers = {}
        if include_headers:
            token = self._get_auth_header_value(self._auth.access_token)
            headers = {HEADER_AUTHORIZATION: token, 'content-type': MIME_JSON, "cache-control": "no-cache"}

        if isinstance(data, dict):
            r = requests.post(self._base_url + url, headers=headers, json=data,  **self._http_request_kwargs)
        else:
            log.error("Unexpected client data type: %s for %s" % (type(data), url))
            raise ClientBadData
        return self.process_server_response(r)

    def _check_token(self):
        if not self._auth.access_token:
            self.get_access_token()
            log.debug("Token is %s, expire date is %s" % (self._auth.access_token, self._auth.token_expires))

        # Token expired, get new token
        elif self._auth.token_expires and self._auth.token_expires < time.time():
            if self._auth.refresh_token:
                self.refresh_access_token()
                log.debug("Token is %s, expire date is %s" % (self._auth.access_token, self._auth.token_expires))
            else:
                raise AuthenticationError("Access Token expired. No refresh token provided.")

    @staticmethod
    def _normalize_and_verify_base_url(u):
        if u.startswith("http://"):
            u = "https://" + u[7:]
        elif not u.startswith("https://"):
            u = "https://" + u
        if not u.endswith("/"):
            u += "/"
        if not re.match(r"^https://[a-z\d]+[-a-z\d.]+[a-z\d][:\d]*/$", u):
            raise ClientBadData
        return u

    def auth(self):
        raise NotImplementedError

    def import_cert(self, request):
        raise NotImplementedError

    # TODO: Need to add service generated CSR implementation
    def request_cert(self, request, zone):
        if not request.csr:
            request.build_csr()
        request_data = {"PolicyDN": self._get_policy_dn(zone),
                        "PKCS10": request.csr,
                        "ObjectName": request.friendly_name,
                        "DisableAutomaticRenewal": "true"}
        if request.origin:
            request_data["Origin"] = request.origin
            ca_origin = {"Name": "Origin", "Value": request.origin}
            if request_data.get("CASpecificAttributes"):
                request_data["CASpecificAttributes"].append(ca_origin)
            else:
                request_data["CASpecificAttributes"] = [ca_origin]

        if request.custom_fields:
            custom_fields_map = {}
            for c_field in request.custom_fields:
                if custom_fields_map.get(c_field.name):
                    custom_fields_map[c_field.name].append(c_field.value)
                else:
                    custom_fields_map[c_field.name] = [c_field.value]

            for key in custom_fields_map:
                custom_field_json = {
                    "Name": key,
                    "Values": custom_fields_map[key]
                }
                if request_data.get("CustomFields"):
                    request_data["CustomFields"].append(custom_field_json)
                else:
                    request_data["CustomFields"] = [custom_field_json]

        status, data = self._post(URLS.CERTIFICATE_REQUESTS, data=request_data)
        if status == HTTPStatus.OK:
            request.id = data['CertificateDN']
            request.cert_guid = data['Guid']
            log.debug("Certificate successfully requested with request id %s." % request.id)
            log.debug("Certificate successfully requested with GUID %s." % request.cert_guid)
            return True

        log.error("Request status is not %s. %s." % HTTPStatus.OK, status)
        raise CertificateRequestError

    def retrieve_cert(self, cert_request):
        log.debug("Getting certificate status for id %s" % cert_request.id)

        retrieve_request = dict(CertificateDN=cert_request.id, Format="base64", IncludeChain='true')

        if cert_request.chain_option == "last":
            retrieve_request['RootFirstOrder'] = 'false'
            retrieve_request['IncludeChain'] = 'true'
        elif cert_request.chain_option == "first":
            retrieve_request['RootFirstOrder'] = 'true'
            retrieve_request['IncludeChain'] = 'true'
        elif cert_request.chain_option == "ignore":
            retrieve_request['IncludeChain'] = 'false'
        else:
            log.error("chain option %s is not valid" % cert_request.chain_option)
            raise ClientBadData

        time_start = time.time()
        while True:
            try:
                status, data = self._post(URLS.CERTIFICATE_RETRIEVE, data=retrieve_request)
            except VenafiError:
                log.debug("Certificate with id %s not found." % cert_request.id)
                status = 0

            if status == HTTPStatus.OK:
                pem64 = data['CertificateData']
                pem = base64.b64decode(pem64)
                return parse_pem(pem.decode(), cert_request.chain_option)
            elif (time.time() - time_start) < cert_request.timeout:
                log.debug("Waiting for certificate...")
                time.sleep(2)
            else:
                raise RetrieveCertificateTimeoutError(
                    'Operation timed out at %d seconds while retrieving certificate with id %s'
                    % (cert_request.timeout, cert_request.id))

    def revoke_cert(self, request):
        if not (request.id or request.thumbprint):
            raise ClientBadData
        d = {
            "Disable": request.disable
        }
        if request.reason:
            d["Reason"] = request.reason
        if request.id:
            d["CertificateDN"] = request.id
        elif request.thumbprint:
            d["Thumbprint"] = request.thumbprint
        else:
            raise ClientBadData
        if request.comments:
            d["Comments"] = request.comments
        status, data = self._post(URLS.CERTIFICATE_REVOKE, data=d)
        if status in (HTTPStatus.OK, HTTPStatus.ACCEPTED):
            return data
        else:
            raise ServerUnexptedBehavior

    def renew_cert(self, request, reuse_key=False):
        if not request.id and not request.thumbprint:
            log.debug("Request id or thumbprint must be specified for TPP")
            raise CertificateRenewError
        if not request.id and request.thumbprint:
            request.id = self.search_by_thumbprint(request.thumbprint)
        if reuse_key:
            log.debug("Trying to renew certificate %s" % request.id)
            status, data = self._post(URLS.CERTIFICATE_RENEW, data={"CertificateDN": request.id})
            if not data['Success']:
                raise CertificateRenewError
            return
        cert = self.retrieve_cert(request)
        cert = x509.load_pem_x509_certificate(cert.cert.encode(), default_backend())
        for a in cert.subject:
            if a.oid == x509.NameOID.COMMON_NAME:
                request.common_name = a.value
            elif a.oid == x509.NameOID.COUNTRY_NAME:
                request.country = a.value
            elif a.oid == x509.NameOID.LOCALITY_NAME:
                request.locality = a.value
            elif a.oid == x509.NameOID.STATE_OR_PROVINCE_NAME:
                request.province = a.value
            elif a.oid == x509.NameOID.ORGANIZATION_NAME:
                request.organization = a.value
            elif a.oid == x509.NameOID.ORGANIZATIONAL_UNIT_NAME:
                request.organizational_unit = a.value
        for e in cert.extensions:
            if e.oid == x509.OID_SUBJECT_ALTERNATIVE_NAME:
                request.san_dns = list([x.value for x in e.value if isinstance(x, x509.DNSName)])
                request.email_addresses = list([x.value for x in e.value if isinstance(x, x509.RFC822Name)])
                request.ip_addresses = list([x.value.exploded for x in e.value if isinstance(x, x509.IPAddress)])
                # remove header bytes from ASN1 encoded UPN field before setting it in the request object
                upns = []
                for x in e.value:
                    if isinstance(x,x509.OtherName):
                        upns.append(x.value[2 : :])
                request.user_principal_names = upns
                request.uniform_resource_identifiers = list([x.value for x in e.value if isinstance(x,x509.UniformResourceIdentifier)])
        if cert.signature_algorithm_oid in (AlgOID.ECDSA_WITH_SHA1, AlgOID.ECDSA_WITH_SHA224, AlgOID.ECDSA_WITH_SHA256,
                                            AlgOID.ECDSA_WITH_SHA384, AlgOID.ECDSA_WITH_SHA512):
            request.key_type = (KeyType.ECDSA, KeyType.ALLOWED_CURVES[0])
        else:
            request.key_type = KeyType(KeyType.RSA, 2048)  # todo: make parsing key size
        if not request.csr:
            request.build_csr()
        status, data = self._post(URLS.CERTIFICATE_RENEW,
                                  data={"CertificateDN": request.id, "PKCS10": request.csr})
        if status == HTTPStatus.OK:
            if "CertificateDN" in data:
                request.id = data['CertificateDN']
            log.debug("Certificate successfully requested with request id %s." % request.id)
            return True

        log.error("Request status is not %s. %s." % HTTPStatus.OK, status)
        raise CertificateRequestError

    @staticmethod
    def _parse_zone_config_to_policy(data):
        # todo: parse over values to regexps (dont forget tests!)
        p = data["Policy"]
        if p["KeyPair"]["KeyAlgorithm"]["Locked"]:
            if p["KeyPair"]["KeyAlgorithm"]["Value"] == "RSA":
                if p["KeyPair"]["KeySize"]["Locked"]:
                    key_types = [KeyType(KeyType.RSA, p["KeyPair"]["KeySize"]["Value"])]
                else:
                    key_types = [KeyType(KeyType.RSA, x) for x in KeyType.ALLOWED_SIZES]
            elif p["KeyPair"]["KeyAlgorithm"]["Value"] == "ECC":
                if p["KeyPair"]["EllipticCurve"]["Locked"]:
                    key_types = [KeyType(KeyType.ECDSA, p["KeyPair"]["EllipticCurve"]["Value"])]
                else:
                    key_types = [KeyType(KeyType.ECDSA, x) for x in KeyType.ALLOWED_CURVES]
            else:
                raise ServerUnexptedBehavior
        else:
            key_types = []
            if p["KeyPair"].get("KeySize", {}).get("Locked"):
                key_types += [KeyType(KeyType.RSA, p["KeyPair"]["KeySize"]["Value"])]
            else:
                key_types += [KeyType(KeyType.RSA, x) for x in KeyType.ALLOWED_SIZES]
            if p["KeyPair"].get("EllipticCurve", {}).get("Locked"):
                key_types += [KeyType(KeyType.ECDSA, p["KeyPair"]["EllipticCurve"]["Value"])]
            else:
                key_types += [KeyType(KeyType.ECDSA, x) for x in KeyType.ALLOWED_CURVES]
        return Policy(key_types=key_types)

    @staticmethod
    def _parse_zone_data_to_object(data):
        s = data["Policy"]["Subject"]
        ou = s['OrganizationalUnit'].get('Values')
        policy = TPPTokenConnection._parse_zone_config_to_policy(data)
        if data["Policy"]["KeyPair"]["KeyAlgorithm"]["Value"] == "RSA":
            key_type = KeyType(KeyType.RSA, data["Policy"]["KeyPair"]["KeySize"]["Value"])
        elif data["Policy"]["KeyPair"]["KeyAlgorithm"]["Value"] == "ECC":
            key_type = KeyType(KeyType.ECDSA, data["Policy"]["KeyPair"]["EllipticCurve"]["Value"])
        else:
            key_type = None

        z = ZoneConfig(
            organization=CertField(s['Organization']['Value'], locked=s['Organization']['Locked']),
            organizational_unit=CertField(ou, locked=s['OrganizationalUnit']['Locked']),
            country=CertField(s['Country']['Value'], locked=s['Country']['Locked']),
            province=CertField(s['State']['Value'], locked=s['State']['Locked']),
            locality=CertField(s['City']['Value'], locked=s['City']['Locked']),
            policy=policy,
            key_type=key_type,
        )
        return z

    def read_zone_conf(self, tag):
        status, data = self._post(URLS.ZONE_CONFIG, {"PolicyDN": self._get_policy_dn(tag)})
        if status != HTTPStatus.OK:
            raise ServerUnexptedBehavior("Server returns %d status on reading zone configuration." % status)
        return self._parse_zone_data_to_object(data)

    @staticmethod
    def _get_policy_dn(zone):
        if zone is None:
            log.error("Bad zone: %s" % zone)
            raise ClientBadData
        if re.match(r"^\\\\VED\\\\Policy", zone):
            return zone
        else:
            if re.match(r"^\\\\", zone):
                return r"\\VED\\Policy" + zone
            else:
                return r"\\VED\\Policy\\" + zone

    def search_by_thumbprint(self, thumbprint):
        """
        :param str thumbprint:
        """
        thumbprint = re.sub(r'[^\dabcdefABCDEF]', "", thumbprint)
        thumbprint = thumbprint.upper()
        status, data = self._get(URLS.CERTIFICATE_SEARCH, params={"Thumbprint": thumbprint})
        if status != HTTPStatus.OK:
            raise ServerUnexptedBehavior

        if not data['Certificates']:
            raise ClientBadData("Certificate not found by thumbprint")
        return data['Certificates'][0]['DN']

    def _get_certificate_details(self, cert_guid):
        status, data = self._get(URLS.CERTIFICATE_SEARCH + cert_guid)
        if status != HTTPStatus.OK:
            raise ServerUnexptedBehavior("")
        return data

    def get_access_token(self, authentication=None):
        """
        Obtains an access token to be used for subsequent api operations.
        """
        if authentication and isinstance(authentication, Authentication):
            self._auth = authentication

        if self._auth.refresh_token:
            return self.refresh_access_token()

        if self._auth.user is None or self._auth.password is None:
            raise ClientBadData("Missing credentials. Cannot request new access token")

        request_data = {
            "username": self._auth.user,
            "password": self._auth.password,
            "client_id": self._auth.client_id,
            "scope": self._auth.scope,
            "state": "",
        }
        status, resp_data = self._post(URLS.AUTHORIZE_TOKEN, request_data, False, False)
        if status != HTTPStatus.OK:
            raise ServerUnexptedBehavior("Server returns %d status on retrieving access token." % status)

        token_info = self._parse_access_token_data_to_object(resp_data)
        self._update_auth(token_info)
        return token_info

    def refresh_access_token(self):
        request_data = {
            "refresh_token": self._auth.refresh_token,
            "client_id": self._auth.client_id,
        }
        status, resp_data = self._post(URLS.REFRESH_TOKEN, request_data, False, False)
        if status != HTTPStatus.OK:
            raise ServerUnexptedBehavior("Server returns %d status on refreshing access token" % status)

        token_info = self._parse_access_token_data_to_object(resp_data)
        self._update_auth(token_info)
        return token_info

    def revoke_access_token(self):
        status, resp_data = self._get(url=URLS.REVOKE_TOKEN, params=None, check_token=False)
        if status != HTTPStatus.OK:
            raise ServerUnexptedBehavior("Server returns %d status on revoking access token" % status)
        return status, resp_data

    def _update_auth(self, token_info):
        if isinstance(token_info, TokenInfo):
            self._auth.access_token = token_info.access_token
            self._auth.refresh_token = token_info.refresh_token
            self._auth.token_expire_date = token_info.expires

    @staticmethod
    def _get_auth_header_value(token):
        return 'Bearer ' + token

    @staticmethod
    def _parse_access_token_data_to_object(data):
        token_info = TokenInfo(
            access_token=data["access_token"],
            expires=data["expires"],
            refresh_token=data["refresh_token"],
        )
        return token_info



#
# Copyright 2020 Venafi, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging as log
import re
from pprint import pprint

from vcert.common import CertField, CommonConnection
from vcert.errors import VenafiError, ServerUnexptedBehavior, ClientBadData
from vcert.http import HTTPStatus
from vcert.policy import RPA, SPA
from vcert.policy.pm_tpp import TPPPolicy, is_service_generated_csr, SetAttrResponse, validate_policy_spec, \
    get_int_value

POLICY_ATTR_CLASS = "X509 Certificate"  # type: str
ROOT_PATH = "\\VED\\Policy\\"  # type: str
POLICY_CLASS = "Policy"  # type: str


class URLS:
    API_TOKEN_URL = "vedauth/"  # type: str  # nosec
    API_BASE_URL = "vedsdk/"  # type: str  # nosec

    AUTHORIZE_TOKEN = API_TOKEN_URL + "authorize/oauth"  # type: str
    REFRESH_TOKEN = API_TOKEN_URL + "authorize/token"  # type: str
    REVOKE_TOKEN = API_TOKEN_URL + "revoke/token"  # type: str

    AUTHORIZE = API_BASE_URL + "authorize/"
    CERTIFICATE_REQUESTS = API_BASE_URL + "certificates/request"
    CERTIFICATE_RETRIEVE = API_BASE_URL + "certificates/retrieve"
    FIND_POLICY = API_BASE_URL + "config/findpolicy"
    CERTIFICATE_REVOKE = API_BASE_URL + "certificates/revoke"
    CERTIFICATE_RENEW = API_BASE_URL + "certificates/renew"
    CERTIFICATE_SEARCH = API_BASE_URL + "certificates/"
    CERTIFICATE_IMPORT = API_BASE_URL + "certificates/import"
    ZONE_CONFIG = API_BASE_URL + "certificates/checkpolicy"
    CONFIG_READ_DN = API_BASE_URL + "Config/ReadDn"

    POLICY_IS_VALID = API_BASE_URL + "config/isvalid"
    POLICY_CREATE = API_BASE_URL + "config/create"
    POLICY_SET_ATTRIBUTE = API_BASE_URL + "config/writepolicy"
    POLICY_CLEAR_ATTRIBUTE = API_BASE_URL + "config/clearpolicyattribute"

    def __init__(self):
        pass


class AbstractTPPConnection(CommonConnection):

    def get_policy(self, zone):
        # get policy spec from name
        policy_name = self._normalize_zone(zone)
        exists = self._policy_exists(policy_name)
        if not exists:
            log.error('The Policy %s does not exist', policy_name)
            raise VenafiError

        status, data = self._post(URLS.ZONE_CONFIG, {"PolicyDN": policy_name})
        if status != HTTPStatus.OK:
            raise ServerUnexptedBehavior("Server returns %d status on reading policy configuration." % status)

        if not ('Policy' in data):
            raise VenafiError("Policy structure not found in response data for [%s] policy", policy_name)
        p = data['Policy']
        tpp_policy = TPPPolicy()

        value = 'Value'
        locked = 'Locked'

        # Contact
        if RPA.TPP_CONTACT in p:
            tpp_policy.contact = [p[RPA.TPP_CONTACT][value]]

        # Approver
        if RPA.TPP_APPROVER in p:
            tpp_policy.approver = [p[RPA.TPP_APPROVER][value]]

        # Prohibited wildcard
        if RPA.TPP_WILDCARDS_ALLOWED in p:
            tpp_policy.wildcards_allowed = p[RPA.TPP_WILDCARDS_ALLOWED]

        # Domain suffix whitelist
        if RPA.TPP_DOMAIN_SUFFIX_WHITELIST in p:
            tpp_policy.domain_suffix_whitelist = p[RPA.TPP_DOMAIN_SUFFIX_WHITELIST]

        # Certification Authority
        if RPA.TPP_CERT_AUTHORITY in p:
            tpp_policy.cert_authority = p[RPA.TPP_CERT_AUTHORITY][value]

        # Management Type
        if RPA.TPP_MANAGEMENT_TYPE in p:
            tpp_policy.management_type = p[RPA.TPP_MANAGEMENT_TYPE][value]

        if not ('Subject' in p):
            raise VenafiError("Subject structure not found in response data for [%s] policy", policy_name)
        subject = p['Subject']

        # Organization
        if RPA.TPP_ORGANIZATION in subject:
            org = subject[RPA.TPP_ORGANIZATION]
            tpp_policy.org = CertField(org[value], org[locked])

        # Organizational Unit
        if RPA.TPP_ORG_UNIT in subject:
            ou = subject[RPA.TPP_ORG_UNIT]
            tpp_policy.org_unit = CertField(ou[value+"s"], ou[locked])

        # City
        if RPA.TPP_CITY in subject:
            city = subject[RPA.TPP_CITY]
            tpp_policy.city = CertField(city[value], city[locked])

        # State
        if RPA.TPP_STATE in subject:
            st = subject[RPA.TPP_STATE]
            tpp_policy.state = CertField(st[value], st[locked])

        # Country
        if RPA.TPP_COUNTRY in subject:
            country = subject[RPA.TPP_COUNTRY]
            tpp_policy.country = CertField(country[value], country[locked])

        if not ('KeyPair' in p):
            raise VenafiError("KeyPair structure not found in response data for [%s] policy", policy_name)
        kp = p['KeyPair']

        # Key Algorithm
        if RPA.TPP_KEY_ALGORITHM in kp:
            tpp_policy.key_algo = CertField(kp[RPA.TPP_KEY_ALGORITHM][value], kp[RPA.TPP_KEY_ALGORITHM][locked])

        # Key Bit Strength
        if RPA.TPP_KEY_SIZE in kp:
            tpp_policy.key_bit_str = CertField(kp[RPA.TPP_KEY_SIZE][value], kp[RPA.TPP_KEY_SIZE][locked])

        # Elliptic Curve
        if RPA.TPP_ELLIPTIC_CURVE in kp:
            tpp_policy.elliptic_curve = CertField(kp[RPA.TPP_ELLIPTIC_CURVE][value], kp[RPA.TPP_ELLIPTIC_CURVE][locked])

        # Manual CSR
        if RPA.TPP_CSR_GENERATION in p:
            bool_val = is_service_generated_csr(p[RPA.TPP_CSR_GENERATION][value])
            tpp_policy.service_generated = CertField(bool_val, p[RPA.TPP_CSR_GENERATION][locked])

        # ALLOWED SANS
        if RPA.TPP_DNS_ALLOWED in p:
            tpp_policy.dns_allowed = p[RPA.TPP_DNS_ALLOWED]
        if RPA.TPP_EMAIL_ALLOWED in p:
            tpp_policy.email_allowed = p[RPA.TPP_EMAIL_ALLOWED]
        if RPA.TPP_IP_ALLOWED in p:
            tpp_policy.ip_allowed = p[RPA.TPP_IP_ALLOWED]
        if RPA.TPP_UPN_ALLOWED in p:
            tpp_policy.upn_allowed = p[RPA.TPP_UPN_ALLOWED]
        if RPA.TPP_URI_ALLOWED in p:
            tpp_policy.uri_allowed = p[RPA.TPP_URI_ALLOWED]

        # Allow Private Key Reuse
        if RPA.TPP_ALLOWED_PRIVATE_KEY_REUSE in p:
            tpp_policy.allow_private_key_reuse = p[RPA.TPP_ALLOWED_PRIVATE_KEY_REUSE]

        # TPP Want Renewal
        # tpp_want_renewal, _ = self._get_policy_attr(policy_name, TPPPolicyAttr.TPP_WANT_RENEWAL)
        # if tpp_want_renewal:
        #     tpp_policy.want_renewal = tpp_want_renewal[0]

        log.info("Building Policy Specification")
        spec = tpp_policy.to_policy_spec()

        return spec

    def set_policy(self, zone, policy_spec):
        """
        :param str zone:
        :param PolicySpecification policy_spec:
        :rtype: PolicySpecification
        """
        validate_policy_spec(policy_spec)
        tpp_policy = TPPPolicy.build_tpp_policy(policy_spec)
        name = self._normalize_zone(zone)
        tpp_policy.name = name

        create_policy = False
        policy_exists = self._policy_exists(name)
        if not policy_exists:
            log.info("Policy [%s] does not exist, validating parent existence", name)
            parent_name = self._get_policy_parent(name)
            if self._policy_exists(parent_name) or parent_name == ROOT_PATH:
                log.info("Parent policy [%s] exists", parent_name)
                create_policy = True
            else:
                raise VenafiError("Parent Policy [%s] does not exist", parent_name)

        # Create the policy if necessary
        if create_policy:
            log.info("Creating Policy [%s]", name)
            policy_request_data = {
                "Class": POLICY_CLASS,
                "ObjectDN": name
            }
            status, resp_data = self._post(URLS.POLICY_CREATE, data=policy_request_data)
            if status != HTTPStatus.OK:
                raise VenafiError("Failed to create policy [%s]. Status %s" % (name, status))

        # Set attributes to policy
        if tpp_policy.contact:
            self._set_policy_attr(name, SPA.TPP_CONTACT, tpp_policy.contact, True)
        if tpp_policy.approver:
            self._set_policy_attr(name, SPA.TPP_APPROVER, tpp_policy.approver, True)

        # Reset all values for existing policy before setting the new ones.
        # This way, values that do not get updated will be inherited from parent.
        if policy_exists:
            self._reset_policy(name)

        if tpp_policy.domain_suffix_whitelist:
            self._set_policy_attr(name, SPA.TPP_DOMAIN_SUFFIX_WHITELIST, tpp_policy.domain_suffix_whitelist, True)
        if tpp_policy.cert_authority:
            self._set_policy_attr(name, SPA.TPP_CERT_AUTHORITY, [tpp_policy.cert_authority], False)
        if tpp_policy.org:
            self._set_policy_attr(name, SPA.TPP_ORGANIZATION, [tpp_policy.org.value], tpp_policy.org.locked)
        if tpp_policy.org_unit:
            self._set_policy_attr(name, SPA.TPP_ORG_UNIT, [tpp_policy.org_unit.value], tpp_policy.org_unit.locked)
        if tpp_policy.city:
            self._set_policy_attr(name, SPA.TPP_CITY, [tpp_policy.city.value], tpp_policy.city.locked)
        if tpp_policy.state:
            self._set_policy_attr(name, SPA.TPP_STATE, [tpp_policy.state.value], tpp_policy.state.locked)
        if tpp_policy.country:
            self._set_policy_attr(name, SPA.TPP_COUNTRY, [tpp_policy.country.value], tpp_policy.state.locked)
        if tpp_policy.key_algo:
            self._set_policy_attr(name, SPA.TPP_KEY_ALGORITHM, [tpp_policy.key_algo.value], tpp_policy.key_algo.locked)
        if tpp_policy.key_bit_str:
            self._set_policy_attr(name, SPA.TPP_KEY_BIT_STR, [tpp_policy.key_bit_str.value],
                                  tpp_policy.key_bit_str.locked)
        if tpp_policy.elliptic_curve:
            self._set_policy_attr(name, SPA.TPP_ELLIPTIC_CURVE, [tpp_policy.elliptic_curve.value],
                                  tpp_policy.elliptic_curve.locked)
        if tpp_policy.management_type:
            self._set_policy_attr(name, SPA.TPP_MANAGEMENT_TYPE, [tpp_policy.management_type.value],
                                  tpp_policy.management_type.locked)

        if tpp_policy.wildcards_allowed is not None:
            int_val = get_int_value(not tpp_policy.wildcards_allowed)
            self._set_policy_attr(name, SPA.TPP_PROHIBIT_WILDCARD, [int_val], False)
        if tpp_policy.service_generated is not None:
            int_val = get_int_value(not tpp_policy.service_generated)
            self._set_policy_attr(name, SPA.TPP_MANUAL_CSR, [int_val], tpp_policy.service_generated.locked)
        if tpp_policy.allow_private_key_reuse is not None:
            int_val = get_int_value(tpp_policy.allow_private_key_reuse)
            self._set_policy_attr(name, SPA.TPP_ALLOWED_PRIVATE_KEY_REUSE, [int_val], True)
        if tpp_policy.want_renewal is not None:
            int_val = get_int_value(tpp_policy.want_renewal)
            self._set_policy_attr(name, SPA.TPP_WANT_RENEWAL, [int_val], True)

        prohibited_sans = tpp_policy.get_prohibited_sans()
        if len(prohibited_sans) > 0:
            self._set_policy_attr(name, SPA.TPP_PROHIBITED_SAN_TYPES, prohibited_sans, False)

        return

    def _policy_exists(self, zone):
        """
        :param str zone:
        :rtype bool:
        """
        req_data = {"ObjectDN": zone}
        status, data = self._post(URLS.POLICY_IS_VALID, data=req_data)

        if status != HTTPStatus.OK:
            raise ServerUnexptedBehavior('Could not complete request. Status %s. %s' % (status, pprint(data)))

        if data['Result'] == 1 and data['Object']['TypeName'] == POLICY_CLASS:
            return True
        elif data['Result'] == 400 and 'Error' in data:
            return False

        log.error("Unknown error while executing. Status: %s.Data: %s." % (status, pprint(data)))
        raise VenafiError

    def _set_policy_attr(self, zone, attr_name, attr_values, locked):
        """
        :param str zone: The policy name
        :param str attr_name: The name of the attribute to be set
        :param any attr_values: The values assigned to the attr_name
        :param bool locked: Whether the attribute should be policy locked
        :rtype: tuple[str, SetAttrResponse]
        """
        data = {
            'Locked': locked,
            'ObjectDN': zone,
            'Class': POLICY_ATTR_CLASS,
            'AttributeName': attr_name,
            'Values': attr_values
        }

        status, response = self._post(URLS.POLICY_SET_ATTRIBUTE, data=data)
        if status != HTTPStatus.OK:
            raise ServerUnexptedBehavior('Got status %s from server', status)

        response = self._parse_attr_response(response)

        if response.error:
            raise VenafiError('Error while setting attribute [%s] in policy [%s]' % (attr_name, zone))

        return status, response

    def _reset_policy_attr(self, zone, attr_name):
        """
        :param str zone:
        :param str attr_name:
        :rtype: tuple[str, SetAttrResponse]
        """
        data = {
            'ObjectDN': zone,
            'Class':  POLICY_ATTR_CLASS,
            'AttributeName': attr_name,
        }

        status, response = self._post(URLS.POLICY_CLEAR_ATTRIBUTE, data=data)
        if status != HTTPStatus.OK:
            raise ServerUnexptedBehavior('Got status %s from server', status)

        response = self._parse_attr_response(response)

        if response.error:
            raise VenafiError('Error while setting attribute [%s] in policy [%s]' % (attr_name, zone))

        return status, response

    def _reset_policy(self, zone):
        self._reset_policy_attr(zone, SPA.TPP_DOMAIN_SUFFIX_WHITELIST)
        self._reset_policy_attr(zone, SPA.TPP_PROHIBIT_WILDCARD)
        self._reset_policy_attr(zone, SPA.TPP_CERT_AUTHORITY)
        self._reset_policy_attr(zone, SPA.TPP_ORGANIZATION)
        self._reset_policy_attr(zone, SPA.TPP_ORG_UNIT)
        self._reset_policy_attr(zone, SPA.TPP_CITY)
        self._reset_policy_attr(zone, SPA.TPP_STATE)
        self._reset_policy_attr(zone, SPA.TPP_COUNTRY)
        self._reset_policy_attr(zone, SPA.TPP_KEY_ALGORITHM)
        self._reset_policy_attr(zone, SPA.TPP_KEY_BIT_STR)
        self._reset_policy_attr(zone, SPA.TPP_ELLIPTIC_CURVE)
        self._reset_policy_attr(zone, SPA.TPP_MANUAL_CSR)
        self._reset_policy_attr(zone, SPA.TPP_PROHIBITED_SAN_TYPES)
        self._reset_policy_attr(zone, SPA.TPP_ALLOWED_PRIVATE_KEY_REUSE)
        self._reset_policy_attr(zone, SPA.TPP_WANT_RENEWAL)
        self._reset_policy_attr(zone, SPA.TPP_MANAGEMENT_TYPE)

    @staticmethod
    def _parse_attr_response(response):
        """
        :param dict response:
        :rtype: SetAttrResponse
        """
        if not response:
            raise VenafiError('Response is empty')

        err = response['Error'] if 'Error' in response else None
        result = response['Result'] if 'Result' in response else None

        return SetAttrResponse(result, err)


    @staticmethod
    def _normalize_zone(zone):
        if zone is None:
            log.error("Bad zone: %s" % zone)
            raise ClientBadData
        if re.match(r"^\\VED\\Policy\\.*", zone):
            return zone
        else:
            if re.match(r"^\\", zone):
                return "\\VED\\Policy" + zone
            else:
                return "\\VED\\Policy\\" + zone

    @staticmethod
    def _get_policy_parent(zone):
        """
        :param str zone:
        """
        if zone is None:
            raise ClientBadData('Zone is empty:')
        index = zone.rindex('\\')
        # Return a substring of zone that starts at 0 and ends at index. Zone here is treated as an slice
        # zone[0:index] returns the same result
        return zone[:index]



#
# Copyright 2019 Venafi, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import (absolute_import, division, generators, unicode_literals, print_function, nested_scopes,
                        with_statement)

import re

_PEM_TYPES = [
    "CERTIFICATE",
    "PRIVATE KEY",
    "ENCRYPTED PRIVATE KEY",
    "RSA PRIVATE KEY",
    "CERTIFICATE REQUEST",
]

_PEM_RE = re.compile(
    "-----BEGIN ("
    + "|".join(_PEM_TYPES)
    + """)-----\r?
.+?\r?
-----END \\1-----\r?\n?""",
    re.DOTALL,
)


def parse_pem(pem_str, order):
    parsed = [
        (match.group(1), (match.group(0)))
        for match in _PEM_RE.finditer(pem_str)
    ]
    certs = []
    key = None
    for p in parsed:
        if p[0] == "CERTIFICATE":
            certs.append(p[1])
        elif p[0].endswith("PRIVATE KEY"):
            key = p[1]
    if order == "last":
        return Certificate(certs[0], certs[1:], key)
    else:
        return Certificate(certs[-1], certs[:-1], key)


class Certificate:
    def __init__(self, cert=None, chain=None, key=None):
        self.cert = cert
        self.chain = chain
        self.key = key

    @property
    def full_chain(self):
        if not self.chain:
            return self.cert
        return self.cert + "\n" + "\n".join(self.chain)




#
# Copyright 2019 Venafi, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .common import CertificateRequest, CommonConnection, RevocationRequest, ZoneConfig, CertField, KeyType, CustomField
from .connection_cloud import CloudConnection
from .connection_tpp import TPPConnection
from .connection_tpp_token import TPPTokenConnection
from .connection_fake import FakeConnection
from .pem import Certificate


def Connection(url=None, token=None, user=None, password=None, fake=False, http_request_kwargs=None):
    """
    Return connection based on credentials list.
    Venafi Platform (TPP) required URL, user, password
    Cloud required token and optional URL
    Fake required no parameters
    :param str url: TPP or Venafi Cloud URL (for Cloud is optional)
    :param str token: Venafi Cloud token
    :param str user: TPP user
    :param str password: TPP password
    :param bool fake: Use fake connection
    :param dict[str, Any] http_request_kwargs: Option for work with untrusted  https certificate (only for TPP).
    :rtype CommonConnection:
    """
    if fake:
        return FakeConnection()
    if url and user and password:
        return TPPConnection(user=user, password=password, url=url, http_request_kwargs=http_request_kwargs)
    if token:
        return CloudConnection(token=token, url=url, http_request_kwargs=http_request_kwargs)
    else:
        raise Exception("Bad credentials list")


def venafi_connection(url=None, api_key=None, user=None, password=None, access_token=None, refresh_token=None,
                      fake=False, http_request_kwargs=None):
    """
    Return connection based on credentials list.
    Venafi Platform (TPP) requires URL and access_token (or user and password for getting a new access_token)
    Cloud requires api_key and optional URL
    Fake requires no parameters
    :param str url: TPP or Venafi Cloud URL (for Cloud is optional)
    :param str api_key: Venafi Cloud API Key
    :param str user: TPP username for getting new tokens
    :param str password: TPP password for getting new tokens
    :param str access_token: TPP access token
    :param str refresh_token: TPP refresh token (optional)
    :param bool fake: Use fake connection
    :param dict[str, Any] http_request_kwargs: Option for specifying trust bundle or to operate insecurely.
    :rtype CommonConnection:
    """
    if fake:
        return FakeConnection()
    if url and (access_token or refresh_token or (user and password)):
        return TPPTokenConnection(url=url, user=user, password=password, access_token=access_token,
                                  refresh_token=refresh_token, http_request_kwargs=http_request_kwargs)
    if api_key:
        return CloudConnection(token=api_key, url=url, http_request_kwargs=http_request_kwargs)
    else:
        raise Exception("Bad credentials list")



#
# Copyright 2021 Venafi, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from vcert.common import Policy as Cit, AppDetails, KeyType
from vcert.errors import VenafiError
from vcert.policy import RPA, DEFAULT_CA
from vcert.policy.policy_spec import Policy, Subject, KeyPair, DefaultSubject, DefaultKeyPair, PolicySpecification, \
    Defaults, SubjectAltNames

supported_rsa_key_sizes = [1024, 2048, 4096]
CA_TYPE_DIGICERT = 'DIGICERT'
CA_TYPE_ENTRUST = 'ENTRUST'
REQUESTER_NAME = "Venafi Cloud Service"
REQUESTER_EMAIL = "no-reply@venafi.cloud"
REQUESTER_PHONE = "801-555-0123"
ALLOW_ALL = ".*"
DEFAULT_MAX_VALID_DAYS = 365
DEFAULT_HASH_ALGORITHM = 'SHA256'


def build_policy_spec(cit, ca_info):
    """
    :param Cit cit:
    :param CertificateAuthorityInfo ca_info:
    :rtype: PolicySpecification
    """
    if not cit:
        raise VenafiError("Certificate issuing template is empty")

    ps = PolicySpecification()
    p = Policy()
    p.wildcard_allowed = is_wildcard_allowed(cit.SubjectCNRegexes)
    if len(cit.SubjectCNRegexes) > 0:
        domains = convert_to_string(cit.SubjectCNRegexes, p.wildcard_allowed)
        p.domains = domains
    else:
        p.domains = None

    if cit.validity_period:
        # getting days in format P#D
        days = cit.validity_period[1:len(cit.validity_period)-1]
        int_value = int(days)
        p.max_valid_days = int_value

    if ca_info:
        template = '%s\\%s\\%s'
        ca = template % (ca_info.ca_type, ca_info.ca_account_key, ca_info.vendor_name)
        p.certificate_authority = ca

    s = Subject()
    create_subject = False
    if len(cit.SubjectORegexes) > 0:
        create_subject = True
        s.orgs = cit.SubjectORegexes
    if len(cit.SubjectOURegexes) > 0:
        create_subject = True
        s.org_units = cit.SubjectOURegexes
    if len(cit.SubjectLRegexes) > 0:
        create_subject = True
        s.localities = cit.SubjectLRegexes
    if len(cit.SubjectSTRegexes) > 0:
        create_subject = True
        s.states = cit.SubjectSTRegexes
    if len(cit.SubjectCRegexes) > 0:
        create_subject = True
        s.countries = cit.SubjectCRegexes

    p.subject = s if create_subject else None

    kp = KeyPair()
    create_kp = False
    if len(cit.key_types) > 0:
        key_types = []
        key_sizes = []
        for allowed_kt in cit.key_types:
            kt = allowed_kt.key_type
            kl = allowed_kt.option
            # Only include one instance of the KeyType
            if kt not in key_types:
                key_types.append(kt)
            key_sizes.append(kl)
        create_kp = True
        kp.key_types = key_types
        kp.rsa_key_sizes = key_sizes

    kp.reuse_allowed = cit.key_reuse
    p.key_pair = kp if create_kp else None

    sans = SubjectAltNames(False, False, False, False, False)
    if cit.SANRegexes:
        sans.dns_allowed = True
    p.subject_alt_names = sans

    ps.policy = p

    rs = cit.recommended_settings
    if rs:
        d = Defaults()
        ds = DefaultSubject()
        create_ds = False
        if rs.subjectOValue:
            ds.org = rs.subjectOValue
            create_ds = True
        if rs.subjectOUValue:
            ds.org_units = [rs.subjectOUValue]
            create_ds = True
        if rs.subjectLValue:
            ds.locality = rs.subjectLValue
            create_ds = True
        if rs.subjectSTValue:
            ds.state = rs.subjectSTValue
            create_ds = True
        if rs.subjectCValue:
            ds.country = rs.subjectCValue
            create_ds = True

        d.subject = ds if create_ds else None

        kt = rs.keyType
        if kt:
            dkp = DefaultKeyPair()
            create_dkp = False
            if kt.key_type:
                dkp.key_type = kt.key_type
                create_dkp = True
            if kt.option:
                dkp.rsa_key_size = kt.option
                create_dkp = True
            d.key_pair = dkp if create_dkp else None

        ps.defaults = d
    return ps


def validate_policy_spec(policy_spec):
    """
    :param PolicySpecification policy_spec:
    """
    default_error_msg = 'Default value does not match with policy values.' \
                        '\nAttribute: %s\nDefault value:%s\nPolicy values:%s'
    # validate policy values
    if policy_spec.policy:
        p = policy_spec.policy

        # validate key pair values
        if policy_spec.policy.key_pair:
            if len(policy_spec.policy.key_pair.key_types) > 1:
                raise VenafiError('Key Type values exceeded. Only one Key Type is allowed by Venafi Cloud')

            if policy_spec.policy.key_pair.key_types \
                    and policy_spec.policy.key_pair.key_types[0].lower() != KeyType.RSA:
                raise VenafiError('Key Type [%s] is not supported by Venafi Cloud' % p.key_pair.key_types[0])

            if len(policy_spec.policy.key_pair.rsa_key_sizes) > 0:
                invalid_value = get_invalid_cloud_rsa_key_size_value(policy_spec.policy.key_pair.rsa_key_sizes)
                if invalid_value:
                    raise VenafiError('The Key Size [%s] is not supported by Venafi Cloud' % invalid_value)

        # validate subject CN and SAN regexes
        if p.subject_alt_names:
            sans = get_sans(policy_spec.policy.subject_alt_names)
            if len(sans) > 0:
                for k, v in sans.items():
                    if v is True and not (k == RPA.TPP_DNS_ALLOWED):
                        raise VenafiError('Subject Alt name [%s] is not allowed by Venafi Cloud' % k)

        # validate default subject values against policy values
        if policy_spec.defaults and policy_spec.defaults.subject and policy_spec.policy.subject:
            ds = policy_spec.defaults.subject
            s = policy_spec.policy.subject

            if ds.org and len(s.orgs) > 0:
                if not is_valid_policy_value(s.orgs, ds.org):
                    raise VenafiError(default_error_msg % ('Organization', ds.org, s.orgs))

            if ds.org_units and len(ds.org_units) > 0 and len(s.org_units) > 0:
                if not member_of(ds.org_units, s.org_units):
                    raise VenafiError(default_error_msg % ('Org Units', ds.org_units, s.org_units))

            if ds.locality and len(s.localities) > 0:
                if not is_valid_policy_value(s.localities, ds.locality):
                    raise VenafiError(default_error_msg % ('Localities', ds.locality, s.localities))

            if ds.state and len(s.states) > 0:
                if not is_valid_policy_value(s.states, ds.state):
                    raise VenafiError(default_error_msg % ('States', ds.state, s.states))

            if ds.country and len(s.countries) > 0:
                if not is_valid_policy_value(s.countries, ds.country):
                    raise VenafiError(default_error_msg % ('Countries', ds.country, s.countries))

        # validate default key pair values against policy values
        if policy_spec.defaults and policy_spec.defaults.key_pair and policy_spec.policy.key_pair:
            dkp = policy_spec.defaults.key_pair
            kp = policy_spec.policy.key_pair

            if dkp.key_type and len(kp.key_types) > 0:
                if dkp.key_type not in kp.key_types:
                    raise VenafiError(default_error_msg % ('Key Types', dkp.key_type, kp.key_types))

            if dkp.rsa_key_size and len(kp.rsa_key_sizes) > 0:
                if dkp.rsa_key_size not in kp.rsa_key_sizes:
                    raise VenafiError(default_error_msg % ('RSA Key Sizes', dkp.rsa_key_size, kp.rsa_key_sizes))

            if dkp.elliptic_curve and len(kp.elliptic_curves) > 0:
                if dkp.elliptic_curve not in kp.elliptic_curves:
                    raise VenafiError(default_error_msg % ('Elliptic Curves', dkp.elliptic_curve, kp.elliptic_curves))

            if dkp.service_generated is not None and kp.service_generated is not None:
                if dkp.service_generated != kp.service_generated:
                    raise VenafiError(
                        default_error_msg % ('Service Generated', dkp.service_generated, kp.service_generated))
    else:
        policy_spec.policy = Policy()

    # validate default values when policy is not defined
    if policy_spec.defaults and policy_spec.defaults.key_pair:
        dkp = policy_spec.defaults.key_pair

        if dkp.key_type and dkp.key_type != "RSA":
            raise VenafiError('Default Key Type [%s] is not supported by Venafi Cloud' % dkp.key_type)

        if dkp.rsa_key_size:
            invalid_value = get_invalid_cloud_rsa_key_size_value([dkp.rsa_key_size])
            if invalid_value:
                raise VenafiError('Default Key Size [%d] is not supported by Venafi Cloud' % invalid_value)


def get_invalid_cloud_rsa_key_size_value(rsa_keys):
    """
    :param list[int] rsa_keys:
    :rtype: int
    """
    for v in rsa_keys:
        if v not in supported_rsa_key_sizes:
            return v
    return None


def get_sans(names):
    """
    :param SubjectAltNames names:
    :rtype: dict[str, bool]
    """
    sans = dict()
    if names.dns_allowed is not None:
        sans[RPA.TPP_DNS_ALLOWED] = names.dns_allowed
    if names.ip_allowed is not None:
        sans[RPA.TPP_IP_ALLOWED] = names.ip_allowed
    if names.email_allowed is not None:
        sans[RPA.TPP_EMAIL_ALLOWED] = names.email_allowed
    if names.upn_allowed is not None:
        sans[RPA.TPP_UPN_ALLOWED] = names.upn_allowed
    if names.uri_allowed is not None:
        sans[RPA.TPP_URI_ALLOWED] = names.uri_allowed

    return sans


def is_valid_policy_value(policy_values, default_value):
    """
    :param list[str] policy_values:
    :param str default_value:
    :rtype: bool
    """
    if len(policy_values) == 1 and policy_values[0] == ALLOW_ALL:
        return True
    return True if default_value in policy_values else False


def member_of(sub_list, collection):
    """
    Tests that all the elements of the sublist are present in the collection

    :param list[str] sub_list:
    :param list[str] collection:
    :rtype: bool
    """
    if len(sub_list) == 1 and sub_list[0] == ALLOW_ALL:
        return True
    return all(x in collection for x in sub_list)


def get_ca_info(ca_name):
    """
    :param str ca_name:
    :rtype: CertificateAuthorityInfo
    """
    data = ca_name.split("\\")
    if len(data) < 3:
        raise VenafiError('Certificate Authority name invalid [%s]' % ca_name)

    return CertificateAuthorityInfo(data[0], data[1], data[2])


def build_cit_request(ps, ca_details):
    """
    :param PolicySpecification ps:
    :param CertificateAuthorityDetails ca_details:
    :rtype: dict
    """
    request = dict()

    if ps.policy and ps.policy.certificate_authority:
        ca_str = ps.policy.certificate_authority
    else:
        ca_str = DEFAULT_CA

    cert_auth = get_ca_info(ca_str)

    request['certificateAuthority'] = cert_auth.ca_type
    request['certificateAuthorityProductOptionId'] = ca_details.product_option_id

    if ps.policy and ps.policy.max_valid_days:
        validity = ps.policy.max_valid_days
    else:
        validity = DEFAULT_MAX_VALID_DAYS

    product = {
        'certificateAuthority': cert_auth.ca_type,
        'productName': cert_auth.vendor_name,
        'validityPeriod': "P%sD" % validity
    }

    if cert_auth.ca_type == CA_TYPE_DIGICERT:
        product['hashAlgorithm'] = DEFAULT_HASH_ALGORITHM
        product['autoRenew'] = False
        product['organizationId'] = ca_details.organization_id
    elif cert_auth.ca_type == CA_TYPE_ENTRUST:
        tracking_data = {
            'certificateAuthority': CA_TYPE_ENTRUST,
            'requesterName': REQUESTER_NAME,
            'requesterEmail': REQUESTER_EMAIL,
            'requesterPhone': REQUESTER_PHONE
        }
        request['trackingData'] = tracking_data

    request['product'] = product

    if ps.policy and len(ps.policy.domains) > 0:
        regex_value = convert_to_regex(ps.policy.domains, ps.policy.wildcard_allowed)
        request['subjectCNRegexes'] = regex_value
        if ps.policy.subject_alt_names and ps.policy.subject_alt_names.dns_allowed is not None:
            if ps.policy.subject_alt_names.dns_allowed:
                request['sanRegexes'] = regex_value
        else:
            request['sanRegexes'] = regex_value
    else:
        request['subjectCNRegexes'] = [ALLOW_ALL]
        request['sanRegexes'] = [ALLOW_ALL]

    if ps.policy and ps.policy.subject and len(ps.policy.subject.orgs) > 0:
        request['subjectORegexes'] = ps.policy.subject.orgs
    else:
        request['subjectORegexes'] = [ALLOW_ALL]

    if ps.policy and ps.policy.subject and len(ps.policy.subject.org_units) > 0:
        request['subjectOURegexes'] = ps.policy.subject.org_units
    else:
        request['subjectOURegexes'] = [ALLOW_ALL]

    if ps.policy and ps.policy.subject and len(ps.policy.subject.localities) > 0:
        request['subjectLRegexes'] = ps.policy.subject.localities
    else:
        request['subjectLRegexes'] = [ALLOW_ALL]

    if ps.policy and ps.policy.subject and len(ps.policy.subject.states) > 0:
        request['subjectSTRegexes'] = ps.policy.subject.states
    else:
        request['subjectSTRegexes'] = [ALLOW_ALL]

    if ps.policy and ps.policy.subject and len(ps.policy.subject.countries) > 0:
        request['subjectCValues'] = ps.policy.subject.countries
    else:
        request['subjectCValues'] = [ALLOW_ALL]

    key_types = dict()
    if ps.policy and ps.policy.key_pair and len(ps.policy.key_pair.key_types) > 0:
        key_types['keyType'] = ps.policy.key_pair.key_types[0].upper()
    else:
        key_types['keyType'] = KeyType.RSA.upper()

    if ps.policy and ps.policy.key_pair and len(ps.policy.key_pair.rsa_key_sizes) > 0:
        key_types['keyLengths'] = ps.policy.key_pair.rsa_key_sizes
    elif ps.defaults and ps.defaults.key_pair and ps.defaults.key_pair.rsa_key_size:
        key_types['keyLengths'] = [ps.defaults.key_pair.rsa_key_size]
    else:
        key_types['keyLengths'] = [2048]

    request['keyTypes'] = [key_types]

    if ps.policy and ps.policy.key_pair and ps.policy.key_pair.reuse_allowed:
        request['keyReuse'] = ps.policy.key_pair.reuse_allowed
    else:
        request['keyReuse'] = False

    r_settings = dict()
    if ps.defaults and ps.defaults.subject:
        if ps.defaults.subject.org:
            r_settings['subjectOValue'] = ps.defaults.subject.org
        if ps.defaults.subject.org_units:
            r_settings['subjectOUValue'] = ps.defaults.subject.org_units[0]
        if ps.defaults.subject.locality:
            r_settings['subjectLValue'] = ps.defaults.subject.locality
        if ps.defaults.subject.state:
            r_settings['subjectSTValue'] = ps.defaults.subject.state
        if ps.defaults.subject.country:
            r_settings['subjectCValue'] = ps.defaults.subject.country

    r_key = dict()
    if ps.defaults and ps.defaults.key_pair:
        if ps.defaults.key_pair.key_type:
            r_key['type'] = ps.defaults.key_pair.key_type
            if ps.defaults.key_pair.rsa_key_size:
                r_key['length'] = ps.defaults.key_pair.rsa_key_size
            else:
                r_key['length'] = 2048

    if r_key:
        r_settings['key'] = r_key

    if r_settings:
        r_settings['keyReuse'] = False
        request['recommendedSettings'] = r_settings

    return request


domain_regex = '[a-z]{1}[a-z0-9.-]*\\.'
domain_regex_wildcard = '[*a-z]{1}[a-z0-9.-]*\\.'


def convert_to_regex(domains, wildcard_allowed):
    """
    :param list[str] domains:
    :param bool wildcard_allowed:
    :rtype: dict
    """
    regex_list = []
    for d in domains:
        current = d.replace('.', '\\.')
        if wildcard_allowed:
            current = domain_regex_wildcard + current
        else:
            current = domain_regex + current
        regex_list.append(current)
    return regex_list


def convert_to_string(regexes, wildcard_allowed):
    """
    :param list[str] regexes:
    :param bool wildcard_allowed:
    :return list[str]:
    """
    pattern = domain_regex_wildcard if wildcard_allowed else domain_regex
    string_list = []
    for r in regexes:
        if r.startswith(pattern):
            r = r.replace(pattern, '')
        r = r.replace('\\.', '.')
        string_list.append(r)
    return string_list


def is_wildcard_allowed(san_regexes):
    """
    :param list[str] san_regexes:
    :rtype: bool
    """
    if not san_regexes:
        return False
    for val in san_regexes:
        if not val.startswith('[*a'):
            return False

    return True


def build_app_update_request(app_details, cit_data):
    """
    :param AppDetails app_details:
    :param dict cit_data:
    :rtype: dict
    """
    app_request = {
        'ownerIdsAndTypes': app_details.owner_ids_and_types,
        'name': app_details.name,
        'description': app_details.description,
        'fqdns': app_details.fq_dns,
        'internalFqdns': app_details.internal_fq_dns,
        'internalIpRanges': app_details.internal_ip_ranges,
        'externalIpRanges': app_details.external_ip_ranges,
        'internalPorts': app_details.internal_ports,
        'fullyQualifiedDomainNames': app_details.fully_qualified_domain_names,
        'ipRanges': app_details.ip_ranges,
        'ports': app_details.ports,
        'organizationalUnitId': app_details.org_unit_id
    }

    cit_map = app_details.cit_alias_id_map

    cit_id, cit_name = get_cit_data_from_response(cit_data)
    cit_map[cit_name] = cit_id

    app_request['certificateIssuingTemplateAliasIdMap'] = cit_map
    return app_request


def build_app_create_request(app_name, user_details, cit_data):
    """

    :param str app_name:
    :param UserDetails user_details:
    :param dict cit_data:
    :rtype: dict
    """
    owner_id = {
        'ownerId': user_details.user.user_id,
        'ownerType': 'USER'
    }
    cit_id, cit_name = get_cit_data_from_response(cit_data)

    app_issuing_template = {
        cit_name: cit_id
    }

    app_request = {
        'ownerIdsAndTypes': [owner_id],
        'name': app_name,
        'certificateIssuingTemplateAliasIdMap': app_issuing_template
    }
    return app_request


def get_cit_data_from_response(data):
    """
    Returns the issuing template id and name from the response after creation

    :param dict data:
    :rtype: (str, str)
    """
    cit_id = None
    cit_name = None
    if 'certificateIssuingTemplates' in data:
        cit_list = data['certificateIssuingTemplates']
        if cit_list and len(cit_list) > 0:
            cit_id = data['certificateIssuingTemplates'][0]['id']
            cit_name = data['certificateIssuingTemplates'][0]['name']
    elif 'id' in data:
        cit_id = data['id']
        cit_name = data['name']

    if cit_name and cit_id:
        return cit_id, cit_name
    else:
        raise VenafiError('Error while creating Application request. CIT name or id not found.')


class CertificateAuthorityInfo:
    def __init__(self, ca_type=None, ca_acc_key=None, vendor_name=None):
        """
        :param str ca_type:
        :param str ca_acc_key:
        :param str vendor_name:
        """
        self.ca_type = ca_type
        self.ca_account_key = ca_acc_key
        self.vendor_name = vendor_name


class CertificateAuthorityDetails:
    def __init__(self, product_option_id=None, organization_id=None):
        """
        :param str product_option_id:
        :param int organization_id:
        """
        self.product_option_id = product_option_id
        self.organization_id = organization_id


class AccountDetails:
    def __init__(self, account, product_options):
        """
        :param Account account:
        :param list[ProductOption] product_options:
        """
        self.account = account
        self.product_options = product_options


class Account:
    def __init__(self, account_id, key, certificate_authority):
        """
        :param str account_id:
        :param str key:
        """
        self.id = account_id
        self.key = key
        self.certificate_authority = certificate_authority


class ProductOption:
    def __init__(self, name, product_id, details=None):
        """
        :param str name:
        :param str product_id:
        :param ProductDetails details:
        """
        self.product_name = name
        self.product_id = product_id
        self.details = details


class ProductDetails:
    def __init__(self, product_template=None):
        """
        :param ProductTemplate product_template:
        """
        self.product_template = product_template


class ProductTemplate:
    def __init__(self, organization_id):
        """
        :param int organization_id:
        """
        self.organization_id = organization_id


class UserDetails:
    def __init__(self, user=None, company=None, api_key=None):
        """
        :param user:
        :param company:
        :param api_key:
        """
        self.user = user
        self.company = company
        self.api_key = api_key


class User:
    def __init__(self, username=None, user_id=None, company_id=None, email=None, user_type=None, user_acc_type=None,
                 status=None, creation_date=None):
        """
        :param str username:
        :param str user_id:
        :param str company_id:
        :param str email:
        :param str user_type:
        :param str user_acc_type:
        :param str status:
        :param str creation_date:
        """
        self.username = username
        self.user_id = user_id
        self.company_id = company_id
        self.email = email
        self.user_type = user_type
        self.user_account_type = user_acc_type
        self.status = status
        self.creation_date = creation_date


class Company:
    def __init__(self, company_id, name, company_type=None, active=None, creation_date=None, domains=None):
        """
        :param str company_id:
        :param str name:
        :param str company_type:
        :param bool active:
        :param str creation_date:
        :param list[str] domains:
        """
        self.company_id = company_id
        self.name = name
        self.company_type = company_type
        self.active = active
        self.creation_date = creation_date
        self.domains = domains


class ApiKey:
    def __init__(self, username=None, api_types=None, api_version=None, apikey_status=None, creation_date=None,
                 validity_start_date=None, validity_end_date=None):
        """
        :param string username:
        :param list[str] api_types:
        :param str api_version:
        :param str apikey_status:
        :param str creation_date:
        :param str validity_start_date:
        :param str validity_end_date:
        """
        self.username = username
        self.api_types = api_types
        self.api_version = api_version
        self.apikey_status = apikey_status
        self.creation_date = creation_date
        self.validity_start_date = validity_start_date
        self.validity_end_date = validity_end_date


def build_user(data):
    """
    :param dict data:
    """
    return User(
        username=data['username'] if 'username' in data else None,
        user_id=data['id'] if 'id' in data else None,
        company_id=data['companyId'] if 'companyId' in data else None,
        email=data['emailAddress'] if 'emailAddress' in data else None,
        user_type=data['userType'] if 'userType' in data else None,
        user_acc_type=data['userAccountType'] if 'userAccountType' in data else None,
        status=data['userStatus'] if 'userStatus' in data else None,
        creation_date=data['creationDate'] if 'creationDate' in data else None
        )


def build_company(data):
    """
    :param dict data:
    """
    return Company(
        company_id=data['id'] if 'id' in data else None,
        name=data['name'] if 'name' in data else None,
        company_type=data['companyType'] if 'companyType' in data else None,
        active=data['active'] if 'active' in data else None,
        creation_date=data['creationDate'] if 'creationDate' in data else None,
        domains=data['domains'] if 'domains' in data else None,
    )


def build_apikey(data):
    """
    :param dict data:
    """
    return ApiKey(
        username=data['username'] if 'username' in data else None,
        api_types=data['apitypes'] if 'apitypes' in data else None,
        api_version=data['apiVersion'] if 'apiVersion' in data else None,
        apikey_status=data['apiKeyStatus'] if 'apiKeyStatus' in data else None,
        creation_date=data['creationDate'] if 'creationDate' in data else None,
        validity_start_date=data['validityStartDate'] if 'validityStartDate' in data else None,
        validity_end_date=data['validityEndDate'] if 'validityEndDate' in data else None,
    )


def build_account_details(data):
    """
    :param dict data:
    :rtype: AccountDetails
    """
    a = 'account'
    if not data or a not in data:
        return

    account = Account(data[a]['id'], data[a]['key'], data[a]['certificateAuthority'])
    po_list = []
    for po in data['productOptions']:
        product = build_product_option(po)
        po_list.append(product)

    ad = AccountDetails(account, po_list)
    return ad


def build_product_option(data):
    """
    :param dict data:
    :rtype: ProductOption
    """
    d = 'productDetails'
    t = 'productTemplate'
    if d in data and t in data[d]:
        org_id = data[d][t]['organizationId'] if 'organizationId' in data[d][t] else None
        p_template = ProductTemplate(org_id)
        p_details = ProductDetails(p_template)
        p_option = ProductOption(data['productName'], data['id'], p_details)
        return p_option



#
# Copyright 2021 Venafi, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging as log
import os

from ruamel.yaml import YAML

from vcert.errors import VenafiParsingError
from vcert.parser.utils import parse_data, load_file, parse_policy_spec


def parse_file(yaml_file_path):
    """
    :param str yaml_file_path: The path of the yaml file to be parsed
    :rtype PolicySpecification:
    """
    if not yaml_file_path:
        log.error('Yaml file path is empty')

    yaml_str = load_file(yaml_file_path)

    return parse(yaml_str)


def parse(yaml_string):
    """
    Parse the yaml string into a Policy Specification object

    :param str yaml_string: the Policy Configuration data in yaml format
    :rtype PolicySpecification:
    """
    if not yaml_string:
        log.error('yaml string is empty')
        raise VenafiParsingError

    yaml = YAML(typ='unsafe')
    data = yaml.load(yaml_string)
    policy = parse_data(data)

    return policy


def serialize(policy_spec, file_path):
    """
    Serializes the policy_spec object into the specified file_path

    :param PolicySpecification policy_spec:
    :param str file_path:
    """
    if not file_path:
        log.error('File path is empty')
    abs_path = os.path.abspath(file_path)
    data = parse_policy_spec(policy_spec)
    f = open(abs_path, 'w')
    yaml = YAML(typ='unsafe')
    yaml.dump(data, f)
    f.close()



#
# Copyright 2021 Venafi, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pprint import pprint

from vcert.common import CertField
from vcert.errors import VenafiError
from vcert.policy import SPA, PolicySpecification, Policy, Subject, DefaultSubject, KeyPair, DefaultKeyPair, Defaults, \
    SubjectAltNames

too_many_error_msg = "attribute [%s] has more than one value"
unsupported_error_msg = "one or more values in attribute [%s] are not supported.\nExpected: %s\nGot: %s"
no_match_error_msg = "default value does not mach with policy value for [%s].\nDefault: %s\nPolicy value:%s"

supported_key_types = ["RSA", "ECDSA"]
supported_rsa_key_sizes = [512, 1024, 2048, 3072, 4096]
supported_elliptic_curves = ["P256", "P384", "P521"]
user_generated_csr = 'UserProvided'
mt_provisioning = "Provisioning"
mt_enrollment = "Enrollment"


class TPPPolicy:
    def __init__(self):
        self.name = ""
        self.contact = None
        self.approver = None
        self.wildcards_allowed = None
        self.domain_suffix_whitelist = None
        self.cert_authority = ""
        # Locked Attr
        self.org = None
        self.org_unit = None
        self.city = None
        self.state = None
        self.country = None
        self.key_algo = None
        self.key_bit_str = None
        self.elliptic_curve = None
        self.service_generated = None
        self.management_type = None
        # End locked Attr

        self.dns_allowed = None
        self.email_allowed = None
        self.ip_allowed = None
        self.upn_allowed = None
        self.uri_allowed = None

        self.allow_private_key_reuse = None
        self.want_renewal = None

    def to_policy_spec(self):
        """
        :rtype: PolicySpecification
        """
        # Building Policy > Subject object
        s = Subject()
        create_subject = False
        ds = DefaultSubject()
        create_default_subject = False

        if self.org:
            if self.org.locked:
                create_subject = True
                s.orgs = [self.org.value]
            else:
                create_default_subject = True
                ds.org = self.org.value

        # Most likely org units value will always be a list , even when only one value exists.
        # So, no list created for the var.
        if self.org_unit:
            if self.org_unit.locked:
                create_subject = True
                s.org_units = self.org_unit.value
            else:
                create_default_subject = True
                ds.org_units = self.org_unit.value

        if self.city:
            if self.city.locked:
                create_subject = True
                s.localities = [self.city.value]
            else:
                create_default_subject = True
                ds.locality = self.city.value

        if self.state:
            if self.state.locked:
                create_subject = True
                s.states = [self.state.value]
            else:
                create_default_subject = True
                ds.state = self.state.value

        if self.country:
            if self.country.locked:
                create_subject = True
                s.countries = [self.country.value]
            else:
                create_default_subject = True
                ds.country = self.country.value

        # Building Policy > KeyPair object
        kp = KeyPair()
        create_kp = False
        dkp = DefaultKeyPair()
        create_default_kp = False

        if self.key_algo:
            if self.key_algo.locked:
                create_kp = True
                kp.key_types = [self.key_algo.value]
            else:
                create_default_kp = True
                dkp.key_type = self.key_algo.value

        if self.key_bit_str:
            int_val = int(self.key_bit_str.value)
            if self.key_bit_str.locked:
                create_kp = True
                kp.rsa_key_sizes = [int_val]
            else:
                create_default_kp = True
                dkp.rsa_key_size = int_val

        if self.elliptic_curve:
            if self.elliptic_curve.locked:
                create_kp = True
                kp.key_types = [self.elliptic_curve.value]
            else:
                create_default_kp = True
                dkp.elliptic_curve = self.elliptic_curve.value

        if self.service_generated:
            if self.service_generated.locked:
                create_kp = True
                kp.service_generated = self.service_generated.value
            else:
                create_default_kp = True
                dkp.service_generated = self.service_generated.value

        kp.reuse_allowed = self.allow_private_key_reuse

        # Building Policy object
        p = Policy()
        p.domains = self.domain_suffix_whitelist
        p.certificate_authority = self.cert_authority
        p.wildcard_allowed = self.wildcards_allowed
        p.auto_installed = self._resolve_management_type()

        if create_subject:
            p.subject = s
        if create_kp:
            p.key_pair = kp

        sans = self._resolve_sans()
        if sans:
            p.subject_alt_names = sans

        # Building Defaults object
        d = Defaults()
        if create_default_subject:
            d.subject = ds
        if create_default_kp:
            d.key_pair = dkp

        # Building Policy Specification
        ps = PolicySpecification()
        ps.users = self.contact
        ps.approvers = self.approver
        ps.policy = p
        ps.defaults = d

        return ps

    @staticmethod
    def build_tpp_policy(ps):
        """
        :param PolicySpecification ps: The PolicySpecification object to convert to TPPPolicy object
        :rtype: TPPPolicy
        """
        tpp_policy = TPPPolicy()
        tpp_policy.contact = ps.users
        tpp_policy.approver = ps.approvers

        policy = ps.policy
        defaults = ps.defaults

        subject = policy.subject if policy else None
        d_subject = defaults.subject if defaults else None

        if policy:
            tpp_policy.domain_suffix_whitelist = policy.domains

        if policy and policy.wildcard_allowed is not None:
            tpp_policy.wildcards_allowed = policy.wildcard_allowed

        if policy and policy.certificate_authority:
            tpp_policy.cert_authority = policy.certificate_authority

        if policy and policy.auto_installed is not None:
            val = get_management_type(policy.auto_installed)
            tpp_policy.management_type = CertField(val, True)
        elif defaults and defaults.auto_installed is not None:
            val = get_management_type(defaults.auto_installed)
            tpp_policy.management_type = CertField(val, False)

        if policy and subject and len(subject.orgs) > 0 and subject.orgs[0]:
            tpp_policy.org = CertField(subject.orgs[0], True)
        elif defaults and d_subject and d_subject.org:
            tpp_policy.org = CertField(d_subject.org, False)

        if policy and subject and len(subject.org_units) > 0 and subject.org_units[0]:
            tpp_policy.org_unit = CertField(subject.org_units[0], True)
        elif defaults and d_subject and len(d_subject.org_units) > 0 and d_subject.org_units[0]:
            tpp_policy.org_unit = CertField(d_subject.org_units[0], False)

        if policy and subject and len(subject.localities) > 0 and subject.localities[0]:
            tpp_policy.city = CertField(subject.localities[0], True)
        elif defaults and d_subject and d_subject.locality:
            tpp_policy.city = CertField(d_subject.locality, False)

        if policy and subject and len(subject.states) > 0 and subject.states[0]:
            tpp_policy.state = CertField(subject.states[0], True)
        elif defaults and d_subject and d_subject.state:
            tpp_policy.state = CertField(d_subject.state, False)

        if policy and subject and len(subject.countries) > 0 and subject.countries[0]:
            tpp_policy.country = CertField(subject.countries[0], True)
        elif defaults and d_subject and d_subject.country:
            tpp_policy.country = CertField(d_subject.country, False)

        kp = policy.key_pair if policy else None
        d_kp = defaults.key_pair if defaults else None

        if policy and kp and len(kp.key_types) > 0 and kp.key_types[0]:
            tpp_policy.key_algo = CertField(kp.key_types[0], True)
        elif defaults and d_kp and d_kp.key_type:
            tpp_policy.key_algo = CertField(d_kp.key_type, False)

        if policy and kp and len(kp.rsa_key_sizes) > 0 and kp.rsa_key_sizes[0]:
            tpp_policy.key_bit_str = CertField(kp.rsa_key_sizes[0], True)
        elif defaults and d_kp and d_kp.rsa_key_size:
            tpp_policy.key_bit_str = CertField(d_kp.rsa_key_size,  False)

        if policy and kp and len(kp.elliptic_curves) > 0 and kp.elliptic_curves[0]:
            tpp_policy.elliptic_curve = CertField(kp.elliptic_curves[0], True)
        elif defaults and d_kp and d_kp.elliptic_curve:
            tpp_policy.elliptic_curve = CertField(d_kp.elliptic_curve, False)

        if policy and kp and kp.service_generated is not None:
            tpp_policy.service_generated = CertField(kp.service_generated, True)
        elif defaults and d_kp and d_kp.service_generated is not None:
            tpp_policy.service_generated = CertField(d_kp.service_generated, False)

        if policy and kp and kp.reuse_allowed is not None:
            bool_val = get_bool_value(kp.reuse_allowed)
            tpp_policy.allow_private_key_reuse = bool_val
            tpp_policy.want_renewal = bool_val

        if policy and policy.subject_alt_names:
            sans = policy.subject_alt_names
            tpp_policy.dns_allowed = sans.dns_allowed
            tpp_policy.email_allowed = sans.email_allowed
            tpp_policy.ip_allowed = sans.ip_allowed
            tpp_policy.upn_allowed = sans.upn_allowed
            tpp_policy.uri_allowed = sans.uri_allowed

        return tpp_policy

    def _resolve_sans(self):
        """
        :rtype SubjectAltNames:
        """
        sans = SubjectAltNames()
        sans.dns_allowed = self.dns_allowed
        sans.ip_allowed = self.ip_allowed
        sans.email_allowed = self.email_allowed
        sans.uri_allowed = self.uri_allowed
        sans.upn_allowed = self.upn_allowed

        return sans

    def get_prohibited_sans(self):
        """
        :rtype: list
        """
        prohibited_sans = []
        if not self.dns_allowed:
            prohibited_sans.append(SPA.TPP_DNS)
        if not self.email_allowed:
            prohibited_sans.append(SPA.TPP_EMAIL)
        if not self.ip_allowed:
            prohibited_sans.append(SPA.TPP_IP)
        if not self.upn_allowed:
            prohibited_sans.append(SPA.TPP_UPN)
        if not self.uri_allowed:
            prohibited_sans.append(SPA.TPP_URI)

        return prohibited_sans

    def _resolve_management_type(self):
        """
        :rtype: bool
        """
        if self.management_type == mt_enrollment:
            return False
        elif self.management_type == mt_provisioning:
            return True


class SetAttrResponse:
    def __init__(self, result=None, error=None):
        """
        :param int result:
        :param str error:
        """
        self.result = result
        self.error = error


def is_service_generated_csr(csr_generation):
    """
    :param str csr_generation:
    :param:
    :rtype: CertField
    """
    if not csr_generation:
        raise VenafiError('csr generation value cannot be empty')

    if csr_generation == user_generated_csr:
        return False
    else:
        return True


def get_bool_value(str_val):
    int_val = int(str_val)
    if int_val == 0:
        return False
    else:
        return True


def get_int_value(bool_val):
    if bool_val:
        return 1
    else:
        return 0


def validate_policy_spec(policy_spec):
    """
    :param PolicySpecification policy_spec:
    :rtype: bool
    """
    if policy_spec.policy:
        validate_policy_subject(policy_spec)
        validate_key_pair(policy_spec)

    validate_default_subject(policy_spec)
    validate_default_key_pair_with_policy_subject(policy_spec)
    validate_default_key_pair(policy_spec)

    d = policy_spec.defaults
    p = policy_spec.policy

    if not d or not p:
        return

    if p.auto_installed is not None:
        if p.auto_installed != d.auto_installed:
            raise VenafiError(no_match_error_msg % ('autoinstalled', d.auto_installed, p.auto_installed))

    return True


def validate_policy_subject(policy_spec):
    """
    :param PolicySpecification policy_spec:
    """
    if not policy_spec.policy.subject:
        raise VenafiError('Subject structure is empty')

    s = policy_spec.policy.subject
    if len(s.orgs) > 1:
        raise VenafiError(too_many_error_msg % 'organizations')
    if len(s.localities) > 1:
        raise VenafiError(too_many_error_msg % 'localities')
    if len(s.states) > 1:
        raise VenafiError(too_many_error_msg % 'states')
    if len(s.countries) > 1:
        raise VenafiError(too_many_error_msg % 'countries')
    # Country values should follow ISO Alpha-2 standard, e.g.: US, MX, CA, FR, etc.
    if len(s.countries[0]) != 2:
        raise VenafiError('country code [%s] does not match ISO Alpha-2 specification' % s.countries[0])


def validate_key_pair(policy_spec):
    """
    :param PolicySpecification policy_spec:
    """
    if not policy_spec.policy.key_pair:
        raise VenafiError('Key Pair structure is empty')

    kp = policy_spec.policy.key_pair

    # validate key algorithm
    if len(kp.key_types) > 1:
        raise VenafiError(too_many_error_msg % 'key types')
    if len(kp.key_types) > 0 and not member_of(kp.key_types, supported_key_types):
        raise VenafiError(unsupported_error_msg % ('key types', pprint(supported_key_types), pprint(kp.key_types)))

    # validate key bit strength
    if len(kp.rsa_key_sizes) > 1:
        raise VenafiError(too_many_error_msg % 'key bit strength')
    if len(kp.rsa_key_sizes) > 0 and not member_of(kp.rsa_key_sizes, supported_rsa_key_sizes):
        raise VenafiError(unsupported_error_msg
                          % ('key bit strength', pprint(supported_rsa_key_sizes), pprint(kp.rsa_key_sizes)))

    # validate elliptic curve
    if len(kp.elliptic_curves) > 1:
        raise VenafiError(too_many_error_msg % 'elliptic curve')
    if len(kp.elliptic_curves) > 0 and not member_of(kp.elliptic_curves, supported_elliptic_curves):
        raise VenafiError(unsupported_error_msg
                          % ('elliptic_curve', pprint(supported_elliptic_curves), pprint(kp.elliptic_curves)))


def validate_default_subject(policy_spec):
    """
    :param PolicySpecification policy_spec:
    """
    if not policy_spec.defaults or not policy_spec.defaults.subject:
        return
    if not policy_spec.policy or not policy_spec.policy.subject:
        return

    s = policy_spec.policy.subject
    ds = policy_spec.defaults.subject

    if s.orgs and s.orgs[0] and ds.org:
        if s.orgs[0] != ds.org:
            raise VenafiError(no_match_error_msg % ('organizations', ds.org, s.orgs[0]))

    if s.org_units and len(s.org_units) > 0 and ds.org_units and len(ds.org_units) > 0:
        if not member_of(ds.org_units, s.org_units):
            raise VenafiError(no_match_error_msg % ('orgUnits', ds.org_units[0], s.org_units[0]))

    if s.localities and s.localities[0] and ds.locality:
        if s.localities[0] != ds.locality:
            raise VenafiError(no_match_error_msg % ('localities', ds.locality, s.localities[0]))

    if s.states and s.states[0] and ds.state:
        if s.states[0] != ds.state:
            raise VenafiError(no_match_error_msg % ('states', ds.state, s.states[0]))

    if s.countries and s.countries and ds.country:
        if s.countries[0] != ds.country:
            raise VenafiError(no_match_error_msg % ('countries', ds.country, s.countries[0]))


def validate_default_key_pair_with_policy_subject(policy_spec):
    """
    :param PolicySpecification policy_spec:
    """
    if not policy_spec.defaults or not policy_spec.defaults.key_pair or not policy_spec.policy \
            or not policy_spec.policy.key_pair:
        return

    kp = policy_spec.policy.key_pair
    dkp = policy_spec.defaults.key_pair

    if kp.key_types and kp.key_types[0] and dkp.key_type:
        if kp.key_types[0] != dkp.key_type:
            raise VenafiError(no_match_error_msg % ('key types', dkp.key_type, kp.key_types[0]))

    if kp.rsa_key_sizes and kp.rsa_key_sizes[0] and dkp.rsa_key_size:
        if kp.rsa_key_sizes[0] != dkp.rsa_key_size:
            raise VenafiError(no_match_error_msg % ('rsa key sizes', dkp.rsa_key_size, kp.rsa_key_sizes[0]))

    if kp.elliptic_curves and kp.elliptic_curves[0] and dkp.elliptic_curve:
        if kp.elliptic_curves[0] != dkp.elliptic_curve:
            raise VenafiError(no_match_error_msg % ('elliptic curves', dkp.elliptic_curve, kp.elliptic_curves[0]))

    if kp.service_generated and dkp.service_generated:
        if kp.service_generated != dkp.service_generated:
            raise VenafiError(no_match_error_msg % ('generation type', dkp.service_generated, kp.service_generated))


def validate_default_key_pair(policy_spec):
    """
    :param PolicySpecification policy_spec:
    """
    if not policy_spec.defaults or not policy_spec.defaults.key_pair:
        return

    dkp = policy_spec.defaults.key_pair

    if dkp.key_type and not member_of([dkp.key_type], supported_key_types):
        raise VenafiError(unsupported_error_msg % ('key type', pprint(supported_key_types), dkp.key_type))

    if dkp.rsa_key_size and not member_of([dkp.rsa_key_size], supported_rsa_key_sizes):
        raise VenafiError(unsupported_error_msg % ('rsa key size', pprint(supported_rsa_key_sizes), dkp.rsa_key_size))

    if dkp.elliptic_curve and not member_of([dkp.elliptic_curve], supported_elliptic_curves):
        raise VenafiError(unsupported_error_msg % ('elliptic curve', pprint(supported_elliptic_curves),
                                                   dkp.elliptic_curve))


def member_of(sub_list, collection):
    """
    Tests that all the elements of the sublist are present in the collection

    :param list collection: The tested values
    :param list sub_list: The member values
    :rtype: bool
    """
    return all(x in collection for x in sub_list)


def get_management_type(autoinstalled):
    """
    :param bool autoinstalled:
    :rtype: str
    """
    if autoinstalled is None:
        return
    elif autoinstalled is True:
        return mt_provisioning
    elif autoinstalled is False:
        return mt_enrollment



#
# Copyright 2021 Venafi, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

FIELD_OWNERS = 'owners'
FIELD_USERS = 'users'
FIELD_USER_ACCESS = 'userAccess'
FIELD_APPROVERS = 'approvers'

FIELD_POLICY = 'policy'
FIELD_DOMAINS = 'domains'
FIELD_WILDCARD_ALLOWED = 'wildcardAllowed'
FIELD_MAX_VALID_DAYS = 'maxValidDays'
FIELD_CERTIFICATE_AUTHORITY = 'certificateAuthority'
FIELD_AUTOINSTALLED = "autoInstalled"

FIELD_SUBJECT = 'subject'
FIELD_ORGS = 'orgs'
FIELD_ORG_UNITS = 'orgUnits'
FIELD_LOCALITIES = 'localities'
FIELD_STATES = 'states'
FIELD_COUNTRIES = 'countries'

FIELD_KEY_PAIR = 'keyPair'
FIELD_KEY_TYPES = 'keyTypes'
FIELD_RSA_KEY_SIZES = 'rsaKeySizes'
FIELD_ELLIPTIC_CURVES = 'ellipticCurves'
FIELD_SERVICE_GENERATED = 'serviceGenerated'
FIELD_REUSE_ALLOWED = 'reuseAllowed'

FIELD_SUBJECT_ALT_NAMES = 'subjectAltNames'
FIELD_DNS_ALLOWED = 'dnsAllowed'
FIELD_IP_ALLOWED = 'ipAllowed'
FIELD_EMAIL_ALLOWED = 'emailAllowed'
FIELD_URI_ALLOWED = 'uriAllowed'
FIELD_UPN_ALLOWED = 'upnAllowed'

FIELD_DEFAULTS = 'defaults'
FIELD_DEFAULT_DOMAIN = 'domain'
FIELD_DEFAULT_AUTOINSTALLED = FIELD_AUTOINSTALLED

FIELD_DEFAULT_SUBJECT = 'subject'
FIELD_DEFAULT_ORG = 'org'
FIELD_DEFAULT_ORG_UNITS = 'orgUnits'
FIELD_DEFAULT_LOCALITY = 'locality'
FIELD_DEFAULT_STATE = 'state'
FIELD_DEFAULT_COUNTRY = 'country'

FIELD_DEFAULT_KEY_PAIR = 'keyPair'
FIELD_DEFAULT_KEY_TYPE = 'keyType'
FIELD_DEFAULT_RSA_KEY_SIZE = 'rsaKeySize'
FIELD_DEFAULT_ELLIPTIC_CURVE = 'ellipticCurve'
FIELD_DEFAULT_SERVICE_GENERATED = 'serviceGenerated'



#
# Copyright 2021 Venafi, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

DEFAULT_CA = "BUILTIN\\Built-In CA\\Default Product"


class PolicySpecification:
    def __init__(self, owners=None, users=None, user_access=None, approvers=None, policy=None, defaults=None):
        """
        :param list[str] owners:
        :param list[str] users:
        :param str user_access:
        :param list[str] approvers:
        :param Policy policy:
        :param Defaults defaults:
        """
        self.owners = owners if owners else []
        self.users = users if users else []
        self.user_access = user_access
        self.approvers = approvers if approvers else []
        self.policy = policy
        self.defaults = defaults


class Policy:
    def __init__(self, domains=None, wildcard_allowed=None, max_valid_days=None, cert_auth=None, subject=None,
                 key_pair=None, subject_alt_names=None, auto_installed=None):
        """
        :param list[str] domains:
        :param bool wildcard_allowed:
        :param int max_valid_days:
        :param str cert_auth:
        :param Subject subject:
        :param KeyPair key_pair:
        :param SubjectAltNames subject_alt_names:
        :param bool auto_installed:
        """
        self.domains = domains if domains else []
        self.wildcard_allowed = wildcard_allowed
        self.max_valid_days = max_valid_days
        self.certificate_authority = cert_auth if cert_auth else DEFAULT_CA
        self.subject = subject
        self.key_pair = key_pair
        self.subject_alt_names = subject_alt_names
        self.auto_installed = auto_installed


class Subject:
    def __init__(self, orgs=None, org_units=None, localities=None, states=None, countries=None):
        """
        :param list[str] orgs:
        :param list[str] org_units:
        :param list[str] localities:
        :param list[str] states:
        :param list[str] countries:
        """
        self.orgs = orgs if orgs else []
        self.org_units = org_units if org_units else []
        self.localities = localities if localities else []
        self.states = states if states else []
        self.countries = countries if countries else []


class KeyPair:
    def __init__(self, key_types=None, rsa_key_sizes=None, elliptic_curves=None, service_generated=None,
                 reuse_allowed=None):
        """
        :param list[str] key_types:
        :param list[int] rsa_key_sizes:
        :param list[str] elliptic_curves:
        :param bool service_generated:
        :param bool reuse_allowed:
        """
        self.key_types = key_types if key_types else []
        self.rsa_key_sizes = rsa_key_sizes if rsa_key_sizes else []
        self.elliptic_curves = elliptic_curves if elliptic_curves else []
        self.service_generated = service_generated
        self.reuse_allowed = reuse_allowed


class SubjectAltNames:
    def __init__(self, dns_allowed=None, email_allowed=None, ip_allowed=None, upn_allowed=None, uri_allowed=None,):
        """
        :param bool dns_allowed:
        :param bool email_allowed:
        :param bool ip_allowed:
        :param bool upn_allowed:
        :param bool uri_allowed:
        """
        self.dns_allowed = dns_allowed
        self.email_allowed = email_allowed
        self.ip_allowed = ip_allowed
        self.upn_allowed = upn_allowed
        self.uri_allowed = uri_allowed


class Defaults:
    def __init__(self, d_domain=None, d_subject=None, d_key_pair=None, auto_installed=None):
        """
        :param str d_domain:
        :param DefaultSubject d_subject:
        :param DefaultKeyPair d_key_pair:
        :param bool auto_installed:
        """
        self.domain = d_domain
        self.subject = d_subject
        self.key_pair = d_key_pair
        self.auto_installed = auto_installed


class DefaultSubject:
    def __init__(self, org=None, org_units=None, locality=None, state=None, country=None):
        """
        :param str org:
        :param list[str] org_units:
        :param str locality:
        :param str state:
        :param str country:
        """
        self.org = org
        self.org_units = org_units if org_units else []
        self.locality = locality
        self.state = state
        self.country = country


class DefaultKeyPair:
    def __init__(self, key_type=None, rsa_key_size=None, elliptic_curve=None, service_generated=None):
        """
        :param str key_type:
        :param int rsa_key_size:
        :param str elliptic_curve:
        :param bool service_generated:
        """
        self.key_type = key_type
        self.rsa_key_size = rsa_key_size
        self.elliptic_curve = elliptic_curve
        self.service_generated = service_generated



#
# Copyright 2021 Venafi, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .policy_spec import Policy, Subject, KeyPair, SubjectAltNames, Defaults, DefaultSubject, DefaultKeyPair, \
    PolicySpecification, DEFAULT_CA


class CommonPA:
    def __init__(self):
        pass

    TPP_CONTACT = "Contact"
    TPP_APPROVER = "Approver"
    TPP_ORGANIZATION = "Organization"
    TPP_CITY = "City"
    TPP_STATE = "State"
    TPP_COUNTRY = "Country"


# tpp read policy attributes - RPA
class RPA(CommonPA):
    def __init__(self):
        pass

    TPP_CERT_AUTHORITY = "CertificateAuthority"  # "Certificate Authority"
    TPP_WILDCARDS_ALLOWED = "WildcardsAllowed"
    TPP_DOMAIN_SUFFIX_WHITELIST = "WhitelistedDomains"  # "Domain Suffix Whitelist"
    TPP_ORG_UNIT = 'OrganizationalUnit'  # "Organizational Unit"
    TPP_KEY_ALGORITHM = 'KeyAlgorithm'  # "Key Algorithm"
    TPP_KEY_SIZE = 'KeySize'
    TPP_ELLIPTIC_CURVE = 'EllipticCurve'  # "Elliptic Curve"
    TPP_CSR_GENERATION = 'CsrGeneration'
    TPP_ALLOWED_PRIVATE_KEY_REUSE = 'PrivateKeyReuseAllowed'  # "Allow Private Key Reuse"
    TPP_WANT_RENEWAL = "Want Renewal"
    TPP_MANAGEMENT_TYPE = "ManagementType"

    TPP_DNS_ALLOWED = "SubjAltNameDnsAllowed"
    TPP_EMAIL_ALLOWED = "SubjAltNameEmailAllowed"
    TPP_IP_ALLOWED = "SubjAltNameIpAllowed"
    TPP_UPN_ALLOWED = "SubjAltNameUpnAllowed"
    TPP_URI_ALLOWED = "SubjAltNameUriAllowed"


# tpp set policy attributes - SPA
class SPA(CommonPA):
    def __init__(self):
        pass

    TPP_CERT_AUTHORITY = "Certificate Authority"
    TPP_PROHIBIT_WILDCARD = "Prohibit Wildcard"
    TPP_DOMAIN_SUFFIX_WHITELIST = "Domain Suffix Whitelist"
    TPP_ORG_UNIT = "Organizational Unit"
    TPP_KEY_ALGORITHM = "Key Algorithm"
    TPP_KEY_BIT_STR = "Key Bit Strength"
    TPP_ELLIPTIC_CURVE = "Elliptic Curve"
    TPP_MANUAL_CSR = "Manual Csr"
    TPP_PROHIBITED_SAN_TYPES = "Prohibited SAN Types"
    TPP_ALLOWED_PRIVATE_KEY_REUSE = "Allow Private Key Reuse"
    TPP_WANT_RENEWAL = "Want Renewal"
    TPP_MANAGEMENT_TYPE = "Management Type"

    TPP_DNS = "DNS"
    TPP_EMAIL = "Email"
    TPP_IP = "IP"
    TPP_UPN = "UPN"
    TPP_URI = "URI"



#
# Copyright 2019 Venafi, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import absolute_import, division, generators, unicode_literals, print_function, nested_scopes, \
    with_statement

import datetime
import logging as log
import time
import uuid

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization

from .common import (ZoneConfig, CommonConnection, Policy, CertField, KeyType)
from .pem import parse_pem

ROOT_CA = b"""-----BEGIN CERTIFICATE-----
MIID1TCCAr2gAwIBAgIJAIOVTvMIMD7OMA0GCSqGSIb3DQEBCwUAMIGAMQswCQYD
VQQGEwJVUzENMAsGA1UECAwEVXRhaDEXMBUGA1UEBwwOU2FsdCBMYWtlIENpdHkx
DzANBgNVBAoMBlZlbmFmaTEbMBkGA1UECwwSTk9UIEZPUiBQUk9EVUNUSU9OMRsw
GQYDVQQDDBJWQ2VydCBUZXN0IE1vZGUgQ0EwHhcNMTgwMzI3MTAyNTI5WhcNMzgw
MzIyMTAyNTI5WjCBgDELMAkGA1UEBhMCVVMxDTALBgNVBAgMBFV0YWgxFzAVBgNV
BAcMDlNhbHQgTGFrZSBDaXR5MQ8wDQYDVQQKDAZWZW5hZmkxGzAZBgNVBAsMEk5P
VCBGT1IgUFJPRFVDVElPTjEbMBkGA1UEAwwSVkNlcnQgVGVzdCBNb2RlIENBMIIB
IjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA0BobDKthxG5SuMfAp2heyDQN
/IL9NTEnFJUUl/CkLEQTSQT68M9US7TCxi+FOizIoev2k4Nkovgk7uM0q94aygbh
cHyTTL64uphHwcClu99ZQ6DIwzDH2gREsLWfj+KXw4bPsne+5tGxv2+0jG2at5or
p/nOQWYD1C1HB6ZQqvP3PypDjou7Uh+Y00bOfXkbYWr8GkX4XAL6UtC0jUnsBEZX
CuwO1BlIIoKNokhOV7Jcb3l/jurjzVWfem+tqwYb/Tkj6MI1YBqt6Yy2EsGsoAv1
E5/IGcjSQnLEqDWhpY0s2fA4o+bAMzyakDFKJoQbF982QhS2fT+d87vQlnMi1QID
AQABo1AwTjAdBgNVHQ4EFgQUzqRFDvLX0mz4AjPb45tLGavm8AcwHwYDVR0jBBgw
FoAUzqRFDvLX0mz4AjPb45tLGavm8AcwDAYDVR0TBAUwAwEB/zANBgkqhkiG9w0B
AQsFAAOCAQEAWbRgS1qUyGMh3ToJ060s5cdoKzyx/ji5pRPXRxrmzzSxP+dlKX7h
AKUgYOV9FU/k2f4C7TeCZSsir20x8fKRg4qs6r8vHTcWnkC6A08SNlT5kjyJl8vt
qQTEsemnyBFis8ZFUfYdmNYqZXuWSb7ZBfNkR7qMVna8A87NyEmTtlTBkZYSTOaB
NRuOli+/6akXg/OW/GfVUD11D413CtZsWNzKaxj1WH88mjBYwQx2pGRzMWHfWBka
f6ZUnA9hhqxO4CHqQWmKPHftbGscwx5yg/J6J7TfG+rYd5ZVVhrr2un2xpOTctjO
lriDCQa4FOwP9/x1OJRXEsSl5YFqBppX5A==
-----END CERTIFICATE-----
"""

ROOT_CA_KEY = b"""-----BEGIN RSA PRIVATE KEY-----
MIIEpAIBAAKCAQEA0BobDKthxG5SuMfAp2heyDQN/IL9NTEnFJUUl/CkLEQTSQT6
8M9US7TCxi+FOizIoev2k4Nkovgk7uM0q94aygbhcHyTTL64uphHwcClu99ZQ6DI
wzDH2gREsLWfj+KXw4bPsne+5tGxv2+0jG2at5orp/nOQWYD1C1HB6ZQqvP3PypD
jou7Uh+Y00bOfXkbYWr8GkX4XAL6UtC0jUnsBEZXCuwO1BlIIoKNokhOV7Jcb3l/
jurjzVWfem+tqwYb/Tkj6MI1YBqt6Yy2EsGsoAv1E5/IGcjSQnLEqDWhpY0s2fA4
o+bAMzyakDFKJoQbF982QhS2fT+d87vQlnMi1QIDAQABAoIBACDBuzhHUeBlrUfA
yaaQWzsQVpNE2y6gShKHVPKFwpHlNVPtIML/H7m6/l3L5SC/I+W5Cts1d4XfoZCo
2wWitHzQkHPwaA9Qhit5BPKOrIfiJF7s1C1FZHAA8/8M180CUflJIzBogPg8Ucpc
fwMLzarQ7cZHIBxTPo8LgX7GwzPlYn/kSlys8w5+gfXCbPIqWHdHgJPIm52ePLZt
0XfTF3Q+HSKzKHpdP/9kJJ3eknM/uEgyOKVPHdkq6U4HeJQq9lmp85X58cyHZOYT
qeLIqgq63x+K3Rgc2EUEOHEbdoEU9aP2fsk9M4AOHf7CpPg7htwN/5m0l4dYpIb+
tcxH+BECgYEA5u9Mtt67y37IIzwQxh80vVZ47LnZMmxKA3AOJOj3Q/fkt36TM+rM
vRFKf6dR6Yolt7JtB6bGrLyGFFbmVDtfDjt9uvseKtG4PUrwjr+ayxICsPPidPlU
hUYh2uu4+m/DK8BGV+PR6/5kwQ2cDF5pdFHECX4VY0uvBnsif1IcBDsCgYEA5rBh
HeKNiUzmfIhP345tZaVM/SFDmAWDs5GZXpCnpMoI2QBg6nU3o6ssPflLjcDjnBrK
VpDlGsTBldX+HhXuEFJzUFASbXXWPqdUyMPzcTQuJWRH+s7Pz94gu/FliNJwmYu2
tsS/PuId4O7dA/Bkhp94sH7OW4iD451xyn4RVC8CgYBtlLu4QrSl+UEKxyPGf2RN
O80ht4aC0LPGMdPkW8+JJwYWtC8xgYcpaB0Lud+6i90d78Kg0NiRetu8pwegjJOs
czpUEXjdJKriGr9PXUgceC1ivjeE/hLHMuI5uYULASGBjzlR1z7zVsGEgeq8S8iK
c4osXvHTLkSdNKzH8bRtpQKBgQDkOZlLHKjULi1VBPKohFr8lcYOJAugacw7Kg+m
u8vvPyXzsekv69mo5Z72jR1PV4aXGPYXIHBYxFGU8Eng7+c/ZKLK0Pz6J/tWrus1
WI8O7wW8XnLL0jFMQED4T0EZVMCrV8rjFNDz4HaY4xfrXrfFbB3V1w5HBk8dL9W0
9HYmZwKBgQCN6xAb82gwFM0H1w4iu6MnA2LyLc/19xn8khgNynW3cUznvyKGQuQo
ZEU0fw9VRRyQVwUwjAaLbIuME4cKhGjcJUvGPLftNamTlFS/TvtE1fwauGBXYc5o
wWh1aVElz5xMF+SnGUCW7t02dvhK0i29mOfx/eG5jkSm33NvVBq/IA==
-----END RSA PRIVATE KEY-----
"""


def fake_user(email=None):
    fake_user_email = email or "test@example.com"
    fake_user_uuid = str(uuid.uuid4())
    fake_company_uuid = str(uuid.uuid4())
    company_domains = ['auth-demo.com', 'example.com']
    fake_company = 'Example Inc.'
    f = {'user': {'username': fake_user_email, 'id': fake_user_uuid, 'companyId': fake_company_uuid,
                  'firstname': 'John', 'lastname': 'Doe', 'emailAddress': fake_user_email, 'userType': 'EXTERNAL',
                  'userAccountType': 'WEB_UI', 'userStatus': 'ACTIVE', 'roles': ['ADMIN'],
                  'firstLoginDate': '2018-11-27T14:24:37.136+0000', 'creationDate': '2018-11-27T14:24:05.455+0000'},
         'company': {'id': fake_company_uuid, 'name': fake_company, 'companyType': 'TPP_CUSTOMER', 'active': True,
                     'creationDate': '2017-04-16T16:49:51.000+0000', 'domains': company_domains},
         'apiKey': {'userId': fake_user_uuid, 'username': fake_user_email, 'companyId': fake_company_uuid,
                    'apiVersion': 'ALL', 'apiKeyStatus': 'ACTIVE', 'creationDate': '2018-11-27T14:24:05.455+0000',
                    'validityStartDate': '2018-11-27T14:24:05.455+0000',
                    'validityEndDate': '2119-05-26T14:24:05.455+0000'}}
    return f


class FakeConnection(CommonConnection):
    def __init__(self, *args, **kwargs):
        self.status = "200"
        self._base_url = "fake"

    def __str__(self):
        return "[Fake]"

    def auth(self):
        return fake_user()

    def register(self, email):
        return fake_user(email)

    def request_cert(self, request, zone):
        if not request.csr:
            request.build_csr()
        data = {"PolicyDN": zone,
                "PKCS10": request.csr,
                "ObjectName": request.friendly_name,
                "CertificateDN": request.friendly_name,
                "DisableAutomaticRenewal": "true"}
        request.id = data['CertificateDN']
        log.debug("Certificate sucessfully requested with request id %s." % request.id)
        return request

    def read_zone_conf(self, tag):
        policy = Policy()
        policy.key_types = list([KeyType(KeyType.RSA, x) for x in KeyType.ALLOWED_SIZES])
        z = ZoneConfig(
            organization=CertField(""),
            organizational_unit=CertField(""),
            country=CertField(""),
            province=CertField(""),
            locality=CertField(""),
            policy=policy,
            key_type=policy.key_types[0],
        )
        return z

    def retrieve_cert(self, certificate_request):
        log.debug("Getting certificate status for id %s" % certificate_request.id)

        time.sleep(0.1)
        certificate_request._public_key_from_private()
        csr = x509.load_pem_x509_csr(certificate_request.csr.encode(), default_backend())

        root_ca_certificate = x509.load_pem_x509_certificate(ROOT_CA, default_backend())
        root_ca_private_key = serialization.load_pem_private_key(ROOT_CA_KEY, password=None,
                                                                 backend=default_backend())


        # cn = x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, certificate_request.common_name)])
        issuer = root_ca_certificate.issuer
        cert = x509.CertificateBuilder().subject_name(
            csr.subject
        ).issuer_name(
            issuer
        ).public_key(
            csr.public_key()
        ).serial_number(
            x509.random_serial_number()
        ).not_valid_before(
            datetime.datetime.utcnow()
        ).not_valid_after(
            # Our certificate will be valid for 10 days
            datetime.datetime.utcnow() + datetime.timedelta(days=10)
        ).add_extension(
            # csr_builder.extensions,
            x509.SubjectAlternativeName([x509.DNSName(u"localhost")]),
            critical=False,
            # Sign our certificate with our private key
        ).sign(root_ca_private_key, hashes.SHA256(), default_backend())

        log.info("This certificate is for test purposes only. Don't use it in production!")
        return parse_pem(cert.public_bytes(serialization.Encoding.PEM).decode(), certificate_request.chain_option)

    def revoke_cert(self, request):
        raise NotImplementedError

    def renew_cert(self, request, reuse_key=False):
        log.debug("Renew is not supported in test mode.")
        raise NotImplementedError



#
# Copyright 2021 Venafi, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging as log
import os

from vcert.parser import *
from vcert.policy import DEFAULT_CA, Subject, KeyPair, SubjectAltNames, Policy, DefaultSubject, DefaultKeyPair, \
    Defaults, PolicySpecification


def load_file(file_path):
    """
    :param str file_path: the path of the file to be loaded
    :rtype str:
    """
    if not file_path:
        log.error('file path is empty')
        raise

    path = os.path.abspath(file_path)
    f = open(path, 'r')
    data = f.read()
    f.close()
    return data


def parse_data(data):
    """
    :param dict data: The data to build the PolicySpecification structure
    :rtype PolicySpecification:
    """
    policy = Policy()
    subject = Subject()
    key_pair = KeyPair()
    subject_alt_names = SubjectAltNames()

    if FIELD_POLICY in data:
        p = data[FIELD_POLICY]
        policy.domains = p[FIELD_DOMAINS] if FIELD_DOMAINS in p else []
        policy.wildcard_allowed = p[FIELD_WILDCARD_ALLOWED] if FIELD_WILDCARD_ALLOWED in p else None
        policy.max_valid_days = p[FIELD_MAX_VALID_DAYS] if FIELD_MAX_VALID_DAYS in p else None
        policy.certificate_authority = p[FIELD_CERTIFICATE_AUTHORITY] if FIELD_CERTIFICATE_AUTHORITY in p \
            else DEFAULT_CA
        policy.auto_installed = p[FIELD_AUTOINSTALLED] if FIELD_AUTOINSTALLED in p else None

        if FIELD_SUBJECT in p:
            s = p[FIELD_SUBJECT]
            subject.orgs = s[FIELD_ORGS] if FIELD_ORGS in s else []
            subject.org_units = s[FIELD_ORG_UNITS] if FIELD_ORG_UNITS in s else []
            subject.localities = s[FIELD_LOCALITIES] if FIELD_LOCALITIES in s else []
            subject.states = s[FIELD_STATES] if FIELD_STATES in s else []
            subject.countries = s[FIELD_COUNTRIES] if FIELD_COUNTRIES in s else []

        if FIELD_KEY_PAIR in p:
            kp = p[FIELD_KEY_PAIR]
            key_pair.key_types = kp[FIELD_KEY_TYPES] if FIELD_KEY_TYPES in kp else []
            key_pair.rsa_key_sizes = kp[FIELD_RSA_KEY_SIZES] if FIELD_RSA_KEY_SIZES in kp else []
            key_pair.elliptic_curves = kp[FIELD_ELLIPTIC_CURVES] if FIELD_ELLIPTIC_CURVES in kp else []
            key_pair.service_generated = kp[FIELD_SERVICE_GENERATED] if FIELD_SERVICE_GENERATED in kp else None
            key_pair.reuse_allowed = kp[FIELD_REUSE_ALLOWED] if FIELD_REUSE_ALLOWED in kp else None

        if FIELD_SUBJECT_ALT_NAMES in p:
            sans = p[FIELD_SUBJECT_ALT_NAMES]
            subject_alt_names.dns_allowed = sans[FIELD_DNS_ALLOWED] if FIELD_DNS_ALLOWED in sans else None
            subject_alt_names.email_allowed = sans[FIELD_EMAIL_ALLOWED] if FIELD_EMAIL_ALLOWED in sans else None
            subject_alt_names.ip_allowed = sans[FIELD_IP_ALLOWED] if FIELD_IP_ALLOWED in sans else None
            subject_alt_names.upn_allowed = sans[FIELD_UPN_ALLOWED] if FIELD_UPN_ALLOWED in sans else None
            subject_alt_names.uri_allowed = sans[FIELD_URI_ALLOWED] if FIELD_URI_ALLOWED in sans else None

    policy.subject = subject
    policy.key_pair = key_pair
    policy.subject_alt_names = subject_alt_names

    defaults = Defaults()
    default_subject = DefaultSubject()
    default_key_pair = DefaultKeyPair()

    if FIELD_DEFAULTS in data:
        d = data[FIELD_DEFAULTS]
        defaults.domain = d[FIELD_DEFAULT_DOMAIN] if FIELD_DEFAULT_DOMAIN in d else None
        defaults.auto_installed = d[FIELD_DEFAULT_AUTOINSTALLED] if FIELD_DEFAULT_AUTOINSTALLED in d else None

        if FIELD_DEFAULT_SUBJECT in d:
            ds = d[FIELD_DEFAULT_SUBJECT]
            default_subject.org = ds[FIELD_DEFAULT_ORG] if FIELD_DEFAULT_ORG in ds else None
            default_subject.org_units = ds[FIELD_DEFAULT_ORG_UNITS] if FIELD_DEFAULT_ORG_UNITS in ds else []
            default_subject.locality = ds[FIELD_DEFAULT_LOCALITY] if FIELD_DEFAULT_LOCALITY in ds else None
            default_subject.state = ds[FIELD_DEFAULT_STATE] if FIELD_DEFAULT_STATE in ds else None
            default_subject.country = ds[FIELD_DEFAULT_COUNTRY] if FIELD_DEFAULT_COUNTRY in ds else None

        if FIELD_DEFAULT_KEY_PAIR in d:
            dkp = d[FIELD_DEFAULT_KEY_PAIR]
            default_key_pair = DefaultKeyPair(dkp[FIELD_DEFAULT_KEY_TYPE],
                                              dkp[FIELD_DEFAULT_RSA_KEY_SIZE],
                                              dkp[FIELD_DEFAULT_ELLIPTIC_CURVE],
                                              dkp[FIELD_DEFAULT_SERVICE_GENERATED])
            default_key_pair.key_type = dkp[FIELD_DEFAULT_KEY_TYPE] if FIELD_DEFAULT_KEY_TYPE in dkp else None
            default_key_pair.rsa_key_size = \
                dkp[FIELD_DEFAULT_RSA_KEY_SIZE] if FIELD_DEFAULT_RSA_KEY_SIZE in dkp else None
            default_key_pair.elliptic_curve = \
                dkp[FIELD_DEFAULT_ELLIPTIC_CURVE] if FIELD_DEFAULT_ELLIPTIC_CURVE in dkp else None
            default_key_pair.service_generated = \
                dkp[FIELD_DEFAULT_SERVICE_GENERATED] if FIELD_DEFAULT_SERVICE_GENERATED in dkp else None

    defaults.subject = default_subject
    defaults.key_pair = default_key_pair

    owners = data[FIELD_OWNERS] if FIELD_OWNERS in data else None
    users = data[FIELD_USERS] if FIELD_USERS in data else None
    user_access = data[FIELD_USER_ACCESS] if FIELD_USER_ACCESS in data else None
    approvers = data[FIELD_APPROVERS] if FIELD_APPROVERS in data else None

    policy_spec = PolicySpecification(owners, users, user_access, approvers, policy, defaults)

    return policy_spec


def parse_policy_spec(policy_spec):
    """
    Returns a dictionary with all the policy specification values. Used for serialization.

    :param PolicySpecification policy_spec:
    :rtype: dict
    """
    s_data = dict()
    kp_data = dict()
    sans_data = dict()
    p_domains = None
    p_wildcard_allowed = None
    p_max_valid_days = None
    p_cert_authority = None

    if policy_spec.policy:
        p_domains = policy_spec.policy.domains
        p_wildcard_allowed = policy_spec.policy.wildcard_allowed
        p_max_valid_days = policy_spec.policy.max_valid_days
        p_cert_authority = policy_spec.policy.certificate_authority

        subject = policy_spec.policy.subject
        if subject:
            s_data = {
                FIELD_ORGS: subject.orgs,
                FIELD_ORG_UNITS: subject.org_units,
                FIELD_LOCALITIES: subject.localities,
                FIELD_STATES: subject.states,
                FIELD_COUNTRIES: subject.countries
            }

        key_pair = policy_spec.policy.key_pair
        if key_pair:
            kp_data = {
                FIELD_KEY_TYPES: key_pair.key_types,
                FIELD_RSA_KEY_SIZES: key_pair.rsa_key_sizes,
                FIELD_ELLIPTIC_CURVES: key_pair.elliptic_curves,
                FIELD_SERVICE_GENERATED: key_pair.service_generated,
                FIELD_REUSE_ALLOWED: key_pair.reuse_allowed
            }

        sans = policy_spec.policy.subject_alt_names
        if sans:
            sans_data = {
                FIELD_DNS_ALLOWED: sans.dns_allowed,
                FIELD_EMAIL_ALLOWED: sans.email_allowed,
                FIELD_IP_ALLOWED: sans.ip_allowed,
                FIELD_UPN_ALLOWED: sans.upn_allowed,
                FIELD_URI_ALLOWED: sans.uri_allowed
            }

    ds_data = dict()
    dkp_data = dict()
    d_domain = None

    if policy_spec.defaults:
        d_domain = policy_spec.defaults.domain

        d_subject = policy_spec.defaults.subject
        if d_subject:
            ds_data = {
                FIELD_DEFAULT_ORG: d_subject.org,
                FIELD_DEFAULT_ORG_UNITS: d_subject.org_units,
                FIELD_DEFAULT_LOCALITY: d_subject.locality,
                FIELD_DEFAULT_STATE: d_subject.state,
                FIELD_DEFAULT_COUNTRY: d_subject.country
            }

        d_key_pair = policy_spec.defaults.key_pair
        if d_key_pair:
            dkp_data = {
                FIELD_DEFAULT_KEY_TYPE: d_key_pair.key_type,
                FIELD_DEFAULT_RSA_KEY_SIZE: d_key_pair.rsa_key_size,
                FIELD_DEFAULT_ELLIPTIC_CURVE: d_key_pair.elliptic_curve,
                FIELD_DEFAULT_SERVICE_GENERATED: d_key_pair.service_generated
            }

    data = {
        FIELD_OWNERS: policy_spec.owners,
        FIELD_USERS: policy_spec.users,
        FIELD_USER_ACCESS: policy_spec.user_access,
        FIELD_APPROVERS: policy_spec.approvers,
        FIELD_POLICY: {
            FIELD_DOMAINS: p_domains,
            FIELD_WILDCARD_ALLOWED: p_wildcard_allowed,
            FIELD_MAX_VALID_DAYS: p_max_valid_days,
            FIELD_CERTIFICATE_AUTHORITY: p_cert_authority,
            FIELD_SUBJECT: s_data,
            FIELD_KEY_PAIR: kp_data,
            FIELD_SUBJECT_ALT_NAMES: sans_data
        },
        FIELD_DEFAULTS: {
            FIELD_DEFAULT_DOMAIN: d_domain,
            FIELD_DEFAULT_SUBJECT: ds_data,
            FIELD_DEFAULT_KEY_PAIR: dkp_data
        }
    }
    new_data = _remove_empty_values(data)
    return new_data


def _remove_empty_values(data):
    """
    Deletes entries whose value is None/empty or equivalent

    :param dict data:
    :rtype: dict
    """
    copy = dict()
    for k, v in data.items():
        if isinstance(v, dict):
            inner_dict = _remove_empty_values(v)
            if len(inner_dict) > 0:
                copy[k] = inner_dict
        elif isinstance(v, list):
            if v is not None and len(v) > 0:
                copy[k] = v
        elif v is not None:
            copy[k] = v

    return copy
